import gc
import logging
from sys import getrefcount
from typing import Tuple

import jpype
import pytest
from pytest import raises

import ixmp
import ixmp.backend.jdbc
from ixmp.backend.jdbc import java
from ixmp.testing import add_random_model_data, bool_param_id, make_dantzig
from ixmp.testing.resource import memory_usage

log = logging.getLogger(__name__)


def test_jvm_warn(recwarn):
    """Test that no warnings are issued on JVM start-up.

    A warning message is emitted e.g. for JPype 0.7 if the 'convertStrings'
    kwarg is not provided to jpype.startJVM.

    NB this function should be in test_core.py, but because pytest executes
    tests in file, then code order, it must be before the call to ix.Platform()
    below.
    """

    # Start the JVM for the first time in the test session
    from ixmp.backend.jdbc import start_jvm

    start_jvm()

    if jpype.__version__ > "0.7":
        # Zero warnings were recorded
        assert len(recwarn) == 0, recwarn.pop().message


def test_close(test_mp_f, capfd):
    """Platform.close_db() doesn't throw needless exceptions."""
    # Use the session-scoped fixture to avoid affecting other tests in this file
    mp = test_mp_f

    # Close once
    mp.close_db()

    # Close again, once already closed
    mp.close_db()

    # With the default log level, WARNING, nothing is printed
    captured = capfd.readouterr()
    assert captured.out == ""

    # With log level INFO, a message is printed
    mp.set_log_level(logging.INFO)
    mp.close_db()
    captured = capfd.readouterr()
    msg = "Database connection could not be closed or was already closed"
    assert msg in captured.out


VE = pytest.mark.xfail(raises=ValueError)


@pytest.mark.parametrize(
    "args, kwargs, expected",
    (
        # Advertised forms
        (
            ("oracle", "url", "user", "pass"),
            dict(),
            dict(driver="oracle", url="url", user="user", password="pass"),
        ),
        (("hsqldb",), dict(url="url"), dict(driver="hsqldb", url="url")),
        # Invalid
        pytest.param(tuple(), dict(), None, marks=VE),
        pytest.param(("not a driver",), dict(), None, marks=VE),
        pytest.param(("oracle", "url", "u", "p", "extra?!"), dict(), None, marks=VE),
        pytest.param(("hsqldb", "path", "extra?!"), dict(), None, marks=VE),
        pytest.param(("oracle", "url", "missing pass"), dict(), None, marks=VE),
        pytest.param(("hsqldb",), dict(), None, marks=VE),
    ),
)
def test_handle_config(args, kwargs, expected):
    """Test :meth:`JDBCBackend.handle_config`."""
    assert expected == ixmp.backend.jdbc.JDBCBackend.handle_config(args, kwargs)


def test_handle_config_path(tmp_path):
    """Test :meth:`JDBCBackend.handle_config` for HyperSQL paths.

    This is separate from :func:`test_handle_config` because the :class:`~pathlib.Path`
    object stored/returned varies across platforms.
    """
    args = ("hsqldb", str(tmp_path))
    kwargs = dict()

    assert dict(
        driver="hsqldb", path=tmp_path
    ) == ixmp.backend.jdbc.JDBCBackend.handle_config(args, kwargs)


def test_exceptions(test_mp):
    """Ensure that Python exceptions are raised for some actions."""
    s = ixmp.Scenario(test_mp, "model name", "scenario name", "new")
    s.init_par("test_exception", [])
    s.commit("")

    with pytest.raises(RuntimeError):
        s.change_scalar("test_exception", 42, unit="kg")


def test_pass_properties():
    ixmp.Platform(
        driver="hsqldb", url="jdbc:hsqldb:mem://ixmptest", user="ixmp", password="ixmp"
    )


def test_invalid_properties_file(test_data_path):
    # HyperSQL creates a file with a .properties suffix for every file-based
    # database, but these files do not contain the information needed to
    # instantiate a database connection
    with pytest.raises(ValueError, match="Config file contains no database URL"):
        ixmp.Platform(dbprops=test_data_path / "hsqldb.properties")


def test_connect_message(capfd, caplog):
    msg = "connected to database 'jdbc:hsqldb:mem://ixmptest' (user: ixmp)..."

    ixmp.Platform(backend="jdbc", driver="hsqldb", url="jdbc:hsqldb:mem://ixmptest")

    # Java code via JPype does not log to the standard Python logger
    assert not any(msg in m for m in caplog.messages)

    # Instead, log messages are printed to stdout
    captured = capfd.readouterr()
    assert msg in captured.out


@pytest.mark.parametrize("arg", [True, False])
def test_cache_arg(arg):
    """Test 'cache' argument, passed to CachingBackend."""
    mp = ixmp.Platform(
        backend="jdbc",
        driver="hsqldb",
        url="jdbc:hsqldb:mem://test_cache_false",
        cache=arg,
    )
    scen = make_dantzig(mp)

    # Maybe put something in the cache
    scen.par("a")

    assert len(mp._backend._cache) == (1 if arg else 0)


def test_read_file(test_mp, tmp_path):
    be = test_mp._backend

    with pytest.raises(NotImplementedError):
        be.read_file(tmp_path / "test.csv", ixmp.ItemType.ALL, filters={})


def test_write_file(test_mp, tmp_path):
    be = test_mp._backend

    with pytest.raises(NotImplementedError):
        be.write_file(tmp_path / "test.csv", ixmp.ItemType.ALL, filters={})


# This variable formerly had 'warns' as the third element in some tuples, to
# test for deprecation warnings.
INIT_PARAMS: Tuple[Tuple, ...] = (
    # Handled in JDBCBackend:
    (
        ["nonexistent.properties"],
        dict(),
        raises,
        ValueError,
        "platform name " r"'nonexistent.properties' not among \['default'",
    ),
    (["nonexistent.properties"], dict(name="default"), raises, TypeError, None),
    # Using the dbtype keyword argument
    (
        [],
        dict(dbtype="HSQLDB"),
        raises,
        TypeError,
        r"JDBCBackend\(\) got an unexpected keyword argument 'dbtype'",
    ),
    # Initialize with driver='oracle' and path
    (
        [],
        dict(backend="jdbc", driver="oracle", path="foo/bar"),
        raises,
        ValueError,
        None,
    ),
    # …with driver='oracle' and no url
    ([], dict(backend="jdbc", driver="oracle"), raises, ValueError, None),
    # …with driver='hsqldb' and no path
    ([], dict(backend="jdbc", driver="hsqldb"), raises, ValueError, None),
    # …with driver='hsqldb' and url
    (
        [],
        dict(backend="jdbc", driver="hsqldb", url="example.com:1234:SCHEMA"),
        raises,
        ValueError,
        None,
    ),
)


@pytest.mark.parametrize("args,kwargs,action,kind,match", INIT_PARAMS)
def test_init(tmp_env, args, kwargs, action, kind, match):
    """Semantics for JDBCBackend.__init__()."""
    with action(kind, match=match):
        ixmp.Platform(*args, **kwargs)


def test_gh_216(test_mp):
    scen = make_dantzig(test_mp)

    filters = dict(i=["seattle", "beijing"])

    # Java code in ixmp_source would raise an exception because 'beijing' is
    # not in set i; but JDBCBackend removes 'beijing' from the filters before
    # calling the underlying method (https://github.com/iiasa/ixmp/issues/216)
    scen.par("a", filters=filters)


@pytest.fixture
def exception_verbose_true():
    """A fixture which ensures JDBCBackend raises verbose exceptions.

    The set value is not disturbed for other tests/code.
    """
    tmp = ixmp.backend.jdbc._EXCEPTION_VERBOSE  # Store current value
    ixmp.backend.jdbc._EXCEPTION_VERBOSE = True  # Ensure True
    yield
    ixmp.backend.jdbc._EXCEPTION_VERBOSE = tmp  # Restore value


def test_verbose_exception(test_mp, exception_verbose_true):
    # Exception stack trace is logged for debugging
    with pytest.raises(RuntimeError) as exc_info:
        ixmp.Scenario(test_mp, model="foo", scenario="bar", version=-1)

    exc_msg = exc_info.value.args[0]
    assert (
        "There exists no Scenario 'foo|bar' "
        "(version: -1)  in the database!" in exc_msg
    )
    assert "at.ac.iiasa.ixmp.database.DbDAO.getRunId" in exc_msg
    assert "at.ac.iiasa.ixmp.Platform.getScenario" in exc_msg


def test_del_ts():
    mp = ixmp.Platform(
        backend="jdbc",
        driver="hsqldb",
        url="jdbc:hsqldb:mem:test_del_ts",
    )

    # Number of Java objects referenced by the JDBCBackend
    N_obj = len(mp._backend.jindex)

    # Create a list of some Scenario objects
    N = 8
    scenarios = [make_dantzig(mp)]
    for i in range(1, N):
        scenarios.append(scenarios[0].clone(scenario=f"clone {i}"))

    # Number of referenced objects has increased by 8
    assert len(mp._backend.jindex) == N_obj + N

    # Pop and free the objects
    for i in range(N):
        s = scenarios.pop(0)

        # The variable 's' is the only reference to this Scenario object
        assert getrefcount(s) - 1 == 1

        # ID of the Scenario object
        s_id = id(s)

        # Underlying Java object
        s_jobj = mp._backend.jindex[s]

        # Now delete the Scenario object
        del s

        # Number of referenced objects decreases by 1
        assert len(mp._backend.jindex) == N_obj + N - (i + 1)
        # ID is no longer in JDBCBackend.jindex
        assert s_id not in mp._backend.jindex

        # s_jobj is the only remaining reference to the Java object
        assert getrefcount(s_jobj) - 1 == 1
        del s_jobj

    # Backend is again empty
    assert len(mp._backend.jindex) == N_obj


# NB coverage is omitted because this test is not included in the standard suite
@pytest.mark.performance
def test_gc_lowmem(request):  # pragma: no cover
    """Test Java-side garbage collection (GC).

    By default, this test limits the JVM memory to 7 MiB. To change this limit, use the
    command-line option --jvm-mem-limit=7.
    """

    # Create a platform with a small memory limit (default 7 MiB)
    mp = ixmp.Platform(
        backend="jdbc",
        driver="hsqldb",
        url="jdbc:hsqldb:mem:test_gc",
        jvmargs=f"-Xmx{request.config.getoption('--jvm-mem-limit', 7)}M",
    )
    # Force Java GC
    jpype.java.lang.System.gc()

    def allocate_scenarios(n):
        for i in range(n):
            scenarios.append(ixmp.Scenario(mp, "foo", f"bar{i}", version="new"))

    scenarios = []
    # Fill *scenarios* with Scenario object until out of memory
    raises(jpype.java.lang.OutOfMemoryError, allocate_scenarios, 100000)
    # Record the maximum number
    max = len(scenarios)

    # Clean up and GC Python and Java memory
    scenarios = []
    gc.collect()
    jpype.java.lang.System.gc()

    # Can allocate *max* scenarios again
    allocate_scenarios(max)
    # …but not twice as many
    raises(jpype.java.lang.OutOfMemoryError, allocate_scenarios, max)


@pytest.fixture(scope="session")
def rc_data_size():  # pragma: no cover
    """Number of data rows for :meth:`test_reload_cycle` and its fixtures."""
    return 5e4


@pytest.fixture(scope="session")
def reload_cycle_scenario(request, tmp_path_factory, rc_data_size):  # pragma: no cover
    """Set up a Platform with *rc_data_size* of  random data."""
    # Command-line option for the JVM memory limit
    kwarg = dict()
    jvm_mem_limit = int(request.config.getoption("--jvm-mem-limit"))
    if jvm_mem_limit > 0:
        kwarg["jvmargs"] = f"-Xmx{jvm_mem_limit}M"

    # Path for this database
    path = tmp_path_factory.mktemp("reload_cycle") / "base"

    # Create the Platform. This should be the first in the process, so the
    # jvmargs are used in :func:`.jdbc.start_jvm`.
    mp = ixmp.Platform(backend="jdbc", driver="hsqldb", path=path, **kwarg)

    s0 = ixmp.Scenario(mp, model="foo", scenario="bar 0", version="new")

    # Add data

    # currently omitted: time series data with *rc_data_size* elements
    # s0.add_timeseries(random_ts_data(rc_data_size))

    # A set named 'random_set' and parameter 'random_par' with *rc_data_size*
    # elements
    add_random_model_data(s0, rc_data_size)

    s0.commit("")

    yield s0


@pytest.mark.performance
@pytest.mark.parametrize("cache", [True, False], ids=bool_param_id("cache"))
@pytest.mark.parametrize("gc", [True, False], ids=bool_param_id("gc"))
@pytest.mark.parametrize("gdx", [True, False], ids=bool_param_id("gdx"))
def test_reload_cycle(
    resource_limit,
    reload_cycle_scenario,
    tmp_path,
    cache,
    gc,
    gdx,
    rc_data_size,
    N_cycles=2,
):  # pragma: no cover
    """Test a cycle of Platform/Scenario reloading.

    This test simulates the usage in the 'runscripts' often used for the
    IIASA MESSAGEix-GLOBIOM global model. Namely:

    1. A large Scenario is created (see the :meth:`reload_cycle_scenario` and
       :meth:`rc_data_size` fixtures).
    2. The Platform instance is discarded, and recreated.
    3. A base Scenario is loaded.
    4. This Scenario is cloned.
    5. The Scenario is solved. (This tests omits this step, but when the `gdx`
       argument is :obj:`True`, the GDX file is written.)
    6. Repeat from (2).

    In order to use this test:

    - Adjust the value in :meth:`rc_data_size`.
    - Adjust the keyword argument `N_cycles`, the minimum number of cycles (
      steps (2) through (5)) required for the test pass.
    - Run by invoking pytest with, e.g.::

        pytest -m performance -k reload_cycle \
          --verbose --log-cli-level=DEBUG -rA \
          --resource-limit=DATA:1024 \
          --jvm-mem-limit=256

      Line by line, these options:

      - Select the current test.
      - Set verbose output, including log messages from the DEBUG level,
        even when the tests pass.
      - Limit Python's resource.RLIMIT_DATA to 1024 MiB.
      - Limit the JVM memory usage to 256 MiB.
    """
    # NB coverage is omitted because this test is not included in the standard
    #    suite

    # Clone reload_cycle_scenario onto a new Platform for this test
    platform_args = dict(
        backend="jdbc",
        driver="hsqldb",
        path=tmp_path / "testdb",
        cache=cache,
        log_level="WARNING",
    )
    mp = ixmp.Platform(**platform_args)
    reload_cycle_scenario.clone(platform=mp)

    # Throw away the reference to mp
    mp = None

    # GC before cycling
    java.System.gc()

    # Set the garbage collection behaviour of JDBCBackend
    ixmp.backend.jdbc._GC_AGGRESSIVE = gc

    pre = memory_usage("setup")

    for i in range(1, N_cycles + 1):
        # New Platform instance; throw away old reference
        mp = ixmp.Platform(**platform_args)

        # Load existing Scenario
        s0 = ixmp.Scenario(mp, model="foo", scenario=f"bar {i-1}", version=1)

        memory_usage(f"pass {i} -- platform instantiated")

        # Clone Scenario
        s1 = s0.clone(scenario=f"bar {i}")

        memory_usage(f"pass {i} -- cloned")

        # Load data into memory
        df_par = s1.par("random_par")
        # commented: omitted
        # df_ts = s1.timeseries()

        memory_usage(f"pass {i} -- data loaded")

        # The variable 's0' is the only reference to this Scenario object
        assert getrefcount(s0) - 1 == 1

        if gdx:
            # Write to file
            mp._backend.write_file(
                tmp_path / "test.gdx",
                ixmp.ItemType.SET | ixmp.ItemType.PAR,
                filters=dict(scenario=s1),
            )

            memory_usage(f"pass {i} -- GDX written")

        # Use, then throw away, references to s0 and data
        assert len(df_par) >= rc_data_size
        s0, df_par = None, None
        # commented: omitted
        # assert len(df_ts) == rc_data_size
        # df_ts = None

        memory_usage(f"pass {i} -- replaced")

        # Throw away reference to mp
        mp = None

    post = memory_usage("post")

    log.info(
        "JVM memory usage gained {:.3f} MiB / cycle".format(
            (post[-2] - pre[-2]) / N_cycles
        )
    )
    log.info(
        "Total memory usage gained {:.3f} MiB / cycle".format(
            (post[0] - pre[0]) / N_cycles
        )
    )

    # Throw away reference to s1
    s1 = None

    memory_usage("shutdown")


def test_docs(test_mp):
    scen = make_dantzig(test_mp)
    # test model docs
    test_mp.set_doc("model", {scen.model: "Dantzig model"})
    assert test_mp.get_doc("model") == {"canning problem": "Dantzig model"}

    # test timeseries variables docs
    gdp = (
        "Gross Domestic Product (GDP) is the monetary value of all "
        "finished goods and services made within a country during "
        "a specific period."
    )
    test_mp.set_doc("timeseries", dict(GDP=gdp))
    assert test_mp.get_doc("timeseries", "GDP") == gdp

    # test bad domain
    ex = raises(ValueError, test_mp.set_doc, "baddomain", {})
    exp = (
        "No such domain: baddomain, existing domains: "
        "scenario, model, region, metadata, timeseries"
    )
    assert ex.value.args[0] == exp


def test_cache_clear(test_mp):
    """Removing set elements causes the cache to be cleared entirely."""
    scen = make_dantzig(test_mp)

    # Load an item so that it is cached
    d0 = scen.par("d")

    # Remove a set element
    scen.check_out()
    scen.remove_set("j", "topeka")

    # The retrieved item content reflects the removal of 'topeka'; not the
    # cached value used to return d0
    assert scen.par("d").shape[0] < d0.shape[0]



"""Test meta functionality of ixmp.Platform and ixmp.Scenario."""

import copy

import pytest

import ixmp
from ixmp.testing import models

SAMPLE_META = {"sample_int": 3, "sample_string": "string_value", "sample_bool": False}
META_ENTRIES = [
    {"sample_int": 3},
    {"sample_string": "string_value"},
    {"sample_bool": False},
    {
        "sample_int": 3,
        "sample_string": "string_value",
        "sample_bool": False,
    },
    {"mixed_category": ["string", 0.01, 2, True]},
]
DANTZIG = models["dantzig"]


@pytest.fixture(scope="function")
def mp(test_mp_f):
    """A test Platform.

    The platform contains one time series with the "dantizg" model & scenario name from
    :data:`testing.models`.

    Unlike the other submodules of :mod:`ixmp.tests.core`, the tests in this file
    generally require a clean platform each time, so this is a function-scoped fixture.
    """
    # Don't call populate_test_platform(), since this is all that's needed
    ts = ixmp.TimeSeries(test_mp_f, **models["dantzig"], version="new")
    ts.commit("")

    yield test_mp_f


@pytest.mark.parametrize("meta", META_ENTRIES)
def test_set_meta_missing_argument(mp, meta):
    with pytest.raises(ValueError):
        mp.set_meta(meta)
    with pytest.raises(ValueError):
        mp.set_meta(meta, model=DANTZIG["model"], version=0)
    with pytest.raises(ValueError):
        mp.set_meta(meta, scenario=DANTZIG["scenario"], version=0)


@pytest.mark.parametrize("meta", META_ENTRIES)
def test_set_get_meta(mp, meta):
    """Assert that storing+retrieving meta yields expected values."""
    mp.set_meta(meta, model=DANTZIG["model"])
    obs = mp.get_meta(model=DANTZIG["model"])
    assert obs == meta


@pytest.mark.parametrize("meta", META_ENTRIES)
def test_unique_meta(mp, meta):
    """
    When setting a meta category on two distinct levels, a uniqueness error is
    expected.
    """
    scenario = ixmp.Scenario(mp, **DANTZIG, version="new")
    scenario.commit("save dummy scenario")
    mp.set_meta(meta, model=DANTZIG["model"])
    expected = (
        r"The meta category .* is already used at another level: "
        r"model canning problem, scenario null, version null"
    )
    with pytest.raises(Exception, match=expected):
        mp.set_meta(meta, **DANTZIG, version=scenario.version)
    scen = ixmp.Scenario(mp, **DANTZIG)
    with pytest.raises(Exception, match=expected):
        scen.set_meta(meta)
    # changing the category value type of an entry should also raise an error
    meta = {"sample_entry": 3}
    mp.set_meta(meta, **DANTZIG)
    meta["sample_entry"] = "test-string"
    expected = (
        r"The meta category .* is already used at another level: "
        r"model canning problem, scenario standard, version null"
    )
    with pytest.raises(Exception, match=expected):
        mp.set_meta(meta, **DANTZIG, version=scenario.version)


@pytest.mark.parametrize("meta", META_ENTRIES)
def test_set_get_meta_equals(mp, meta):
    initial_meta = mp.get_meta(scenario=DANTZIG["scenario"])
    mp.set_meta(meta, model=DANTZIG["model"])
    obs_meta = mp.get_meta(scenario=DANTZIG["scenario"])
    assert obs_meta == initial_meta


@pytest.mark.parametrize("meta", META_ENTRIES)
def test_unique_meta_model_scenario(mp, meta):
    """
    When setting a meta key for a Model, it shouldn't be possible to set it
    for a Model+Scenario then.
    """
    mp.set_meta(meta, model=DANTZIG["model"])
    expected = r"The meta category .* is already used at another level: "
    with pytest.raises(Exception, match=expected):
        mp.set_meta(meta, **DANTZIG)

    # Setting this meta category on a new model should fail too
    dantzig2 = {
        "model": "canning problem 2",
        "scenario": "standard",
    }
    mp.add_model_name(dantzig2["model"])
    expected = r"The meta category .* is already used at another level: "
    with pytest.raises(Exception, match=expected):
        mp.set_meta(meta, **dantzig2)


@pytest.mark.parametrize("meta", META_ENTRIES)
def test_get_meta_strict(mp, meta):
    """
    Set meta indicators on several model/scenario/version levels and test
    the 'strict' parameter of get_meta().
    """
    # set meta on various levels
    model_meta = {"model_int": 3, "model_string": "string_value", "model_bool": False}
    scenario_meta = {
        "scenario_int": 3,
        "scenario_string": "string_value",
        "scenario_bool": False,
    }
    meta2 = {"sample_int2": 3, "sample_string2": "string_value2", "sample_bool2": False}
    meta3 = {
        "sample_int3": 3,
        "sample_string3": "string_value3",
        "sample_bool3": False,
        "mixed3": ["string", 0.01, 2, True],
    }
    meta_scen = {
        "sample_int4": 3,
        "sample_string4": "string_value4",
        "sample_bool4": False,
        "mixed4": ["string", 0.01, 2, True],
    }
    scenario2 = "standard 2"
    model2 = "canning problem 2"
    mp.add_scenario_name(scenario2)
    mp.add_model_name(model2)
    dantzig2 = {
        "model": model2,
        "scenario": "standard",
    }
    dantzig3 = {
        "model": model2,
        "scenario": scenario2,
    }
    mp.set_meta(model_meta, model=DANTZIG["model"])
    mp.set_meta(scenario_meta, scenario=DANTZIG["scenario"])
    mp.set_meta(meta, **DANTZIG)
    mp.set_meta(meta2, **dantzig2)
    mp.set_meta(meta3, **dantzig3)
    scen = ixmp.Scenario(mp, **DANTZIG, version="new")
    scen.commit("save dummy scenario")
    scen.set_meta(meta_scen)

    # Retrieve and validate meta indicators
    # model
    obs1 = mp.get_meta(model=DANTZIG["model"])
    assert obs1 == model_meta
    # scenario
    obs2 = mp.get_meta(scenario=DANTZIG["scenario"], strict=True)
    assert obs2 == scenario_meta
    # model+scenario
    obs3 = mp.get_meta(**DANTZIG)
    exp3 = copy.copy(meta)
    exp3.update(model_meta)
    exp3.update(scenario_meta)
    assert obs3 == exp3
    # model+scenario, strict
    obs3_strict = mp.get_meta(**DANTZIG, strict=True)
    assert obs3_strict == meta
    assert obs3 != obs3_strict

    # second model+scenario combination
    obs4 = mp.get_meta(**dantzig2)
    exp4 = copy.copy(meta2)
    exp4.update(scenario_meta)
    assert obs4 == exp4
    # second model+scenario combination, strict
    obs4_strict = mp.get_meta(**dantzig2, strict=True)
    assert obs4_strict == meta2
    assert obs4 != obs4_strict

    # second model+scenario combination
    obs5 = mp.get_meta(**dantzig3)
    exp5 = copy.copy(meta3)
    assert obs5 == exp5

    # model+scenario+version
    obs6 = mp.get_meta(**DANTZIG, version=scen.version)
    exp6 = copy.copy(meta_scen)
    exp6.update(meta)
    exp6.update(model_meta)
    exp6.update(scenario_meta)
    assert obs6 == exp6
    obs6_strict = mp.get_meta(
        DANTZIG["model"], DANTZIG["scenario"], scen.version, strict=True
    )
    assert obs6_strict == meta_scen


@pytest.mark.parametrize("meta", META_ENTRIES)
def test_unique_meta_scenario(mp, meta):
    """
    When setting a meta key on a specific Scenario run, setting the same key
    on an higher level (Model or Model+Scenario) should fail.
    """
    scen = ixmp.Scenario(mp, **DANTZIG)
    scen.set_meta(meta)
    # add a second scenario and verify that setting+getting Meta works
    scen2 = ixmp.Scenario(mp, **DANTZIG, version="new")
    scen2.commit("save dummy scenario")
    scen2.set_meta(meta)
    assert scen2.get_meta() == scen.get_meta()

    expected = (
        r"The meta category .* is already used at another level: "
        r"model canning problem, scenario standard, "
    )
    with pytest.raises(Exception, match=expected):
        mp.set_meta(meta, **DANTZIG)
    with pytest.raises(Exception, match=expected):
        mp.set_meta(meta, model=DANTZIG["model"])


def test_meta_partial_overwrite(mp):
    meta1 = {
        "sample_string": 3.0,
        "another_string": "string_value",
        "sample_bool": False,
    }
    meta2 = {"sample_string": 5.0, "yet_another_string": "hello", "sample_bool": True}
    scen = ixmp.Scenario(mp, **DANTZIG)
    scen.set_meta(meta1)
    scen.set_meta(meta2)
    expected = copy.copy(meta1)
    expected.update(meta2)
    obs = scen.get_meta()
    assert obs == expected


def test_remove_meta(mp):
    meta = {"sample_int": 3.0, "another_string": "string_value"}
    remove_key = "another_string"
    mp.set_meta(meta, **DANTZIG)
    mp.remove_meta(remove_key, **DANTZIG)
    expected = copy.copy(meta)
    del expected[remove_key]
    obs = mp.get_meta(**DANTZIG)
    assert expected == obs


def test_remove_invalid_meta(mp):
    """
    Removing nonexisting meta entries or None shouldn't result in any meta
    being removed. Providing None should give a ValueError.
    """
    mp.set_meta(SAMPLE_META, **DANTZIG)
    with pytest.raises(ValueError):
        mp.remove_meta(None, **DANTZIG)
    mp.remove_meta("nonexisting_category", **DANTZIG)
    mp.remove_meta([], **DANTZIG)
    obs = mp.get_meta(**DANTZIG)
    assert obs == SAMPLE_META


def test_set_and_remove_meta_scenario(mp):
    """
    Test partial overwriting and meta deletion on scenario level.
    """
    meta1 = {"sample_string": 3.0, "another_string": "string_value"}
    meta2 = {"sample_string": 5.0, "yet_another_string": "hello"}
    remove_key = "another_string"

    scen = ixmp.Scenario(mp, **DANTZIG)
    scen.set_meta(meta1)
    scen.set_meta(meta2)
    expected = copy.copy(meta1)
    expected.update(meta2)
    obs = scen.get_meta()
    assert expected == obs

    scen.remove_meta(remove_key)
    del expected[remove_key]
    obs = scen.get_meta()
    assert obs == expected


def test_scenario_delete_meta_warning(mp):
    """
    Scenario.delete_meta works but raises a deprecation warning.

    This test can be removed once Scenario.delete_meta is removed.
    """
    scen = ixmp.Scenario(mp, **DANTZIG)
    meta = {"sample_int": 3, "sample_string": "string_value"}
    remove_key = "sample_string"

    scen.set_meta(meta)
    with pytest.warns(DeprecationWarning):
        scen.delete_meta(remove_key)
    expected = copy.copy(meta)
    del expected[remove_key]
    obs = scen.get_meta()
    assert obs == expected


def test_meta_arguments(mp):
    """Set scenario meta with key-value arguments"""
    meta = {"sample_int": 3}
    scen = ixmp.Scenario(mp, **DANTZIG)
    scen.set_meta(meta)
    # add a second scenario and verify that setting Meta for it works
    scen2 = ixmp.Scenario(mp, **DANTZIG, version="new")
    scen2.commit("save dummy scenario")
    scen2.set_meta(*meta.popitem())
    assert scen.get_meta() == scen2.get_meta()


def test_update_meta_lists(mp):
    """Set metadata categories having list/array values."""
    SAMPLE_META = {"list_category": ["a", "b", "c"]}
    mp.set_meta(SAMPLE_META, model=DANTZIG["model"])
    obs = mp.get_meta(model=DANTZIG["model"])
    assert obs == SAMPLE_META
    # try updating meta
    SAMPLE_META = {"list_category": ["a", "e", "f"]}
    mp.set_meta(SAMPLE_META, model=DANTZIG["model"])
    obs = mp.get_meta(model=DANTZIG["model"])
    assert obs == SAMPLE_META


def test_meta_mixed_list(mp):
    """Set metadata categories having list/array values."""
    meta = {"mixed_category": ["string", 0.01, True]}
    mp.set_meta(meta, model=DANTZIG["model"])
    obs = mp.get_meta(model=DANTZIG["model"])
    assert obs == meta



from pathlib import Path

import pytest

from ixmp.backend import ItemType
from ixmp.backend.base import Backend, CachingBackend
from ixmp.testing import make_dantzig


class BE1(Backend):
    """Incomplete subclass."""


def noop(self, *args, **kwargs):
    pass


class BE2(Backend):
    """Complete subclass."""

    add_model_name = noop
    add_scenario_name = noop
    cat_get_elements = noop
    cat_list = noop
    cat_set_elements = noop
    check_out = noop
    clear_solution = noop
    clone = noop
    commit = noop
    delete = noop
    delete_geo = noop
    delete_item = noop
    delete_meta = noop
    discard_changes = noop
    get = noop
    get_data = noop
    get_doc = noop
    get_geo = noop
    get_meta = noop
    get_model_names = noop
    get_nodes = noop
    get_scenarios = noop
    get_scenario_names = noop
    get_timeslices = noop
    get_units = noop
    has_solution = noop
    init = noop
    init_item = noop
    is_default = noop
    item_delete_elements = noop
    item_get_elements = noop
    item_index = noop
    item_set_elements = noop
    last_update = noop
    list_items = noop
    remove_meta = noop
    run_id = noop
    set_as_default = noop
    set_data = noop
    set_doc = noop
    set_geo = noop
    set_meta = noop
    set_node = noop
    set_timeslice = noop
    set_unit = noop


def test_class():

    # An incomplete Backend subclass can't be instantiated
    with pytest.raises(
        TypeError, match="Can't instantiate abstract class BE1 with abstract methods"
    ):
        BE1()

    # Complete subclass can be instantiated
    be = BE2()

    # Methods with a default implementation can be called
    with pytest.raises(NotImplementedError):
        be.read_file(Path("foo"), ItemType.VAR)

    with pytest.raises(NotImplementedError):
        be.write_file(Path("foo"), ItemType.VAR)


@pytest.mark.parametrize(
    "args, kwargs",
    (
        (tuple(), dict()),
        # Invalid
        pytest.param(("foo",), dict(), marks=pytest.mark.xfail(raises=ValueError)),
        pytest.param(tuple(), dict(bar=""), marks=pytest.mark.xfail(raises=ValueError)),
    ),
)
def test_handle_config(args, kwargs):
    """Test :meth:`JDBCBackend.handle_config`."""
    assert dict() == Backend.handle_config(args, kwargs)


def test_cache_non_hashable():
    filters = {"s": ["foo", 42, object()]}

    # _cache_key() cannot handle non-hashable object()
    # NB exception message contains single quotes on 'object' on Windows/py3.6
    with pytest.raises(
        TypeError, match="Object of type .?object.? is not JSON" " serializable"
    ):
        CachingBackend._cache_key(object(), "par", "p", filters)


def test_cache_del_ts(test_mp):
    """Test CachingBackend.del_ts()."""
    # Since CachingBackend is an abstract class, test it via JDBCBackend
    backend = test_mp._backend
    cache_size_pre = len(backend._cache)

    # Load data, thereby adding to the cache
    s = make_dantzig(test_mp)
    s.par("d")

    # Cache size has increased
    assert len(backend._cache) == cache_size_pre + 1

    # Delete the object; associated cache is freed
    del s

    # Objects were invalidated/removed from cache
    assert len(backend._cache) == cache_size_pre



"""Utilities for testing ixmp.

These include:

- pytest hooks, `fixtures <https://docs.pytest.org/en/latest/fixture.html>`_:

  .. autosummary::
     :nosignatures:

     ixmp_cli
     tmp_env
     test_mp

  …and assertions:

  .. autosummary::
     assert_logs

- Methods for setting up and populating test ixmp databases:

  .. autosummary::
     add_test_data
     create_test_platform
     make_dantzig
     populate_test_platform

- Methods to run and retrieve values from Jupyter notebooks:

  .. autosummary::
     run_notebook
     get_cell_output

"""
import contextlib
import logging
import os
import shutil
from contextlib import contextmanager
from copy import deepcopy
from itertools import chain
from pathlib import Path

import pint
import pytest
from click.testing import CliRunner

from ixmp import cli
from ixmp import config as ixmp_config
from ixmp.core import Platform

from .data import (
    DATA,
    HIST_DF,
    TS_DF,
    add_random_model_data,
    add_test_data,
    make_dantzig,
    models,
    populate_test_platform,
    random_model_data,
    random_ts_data,
)
from .jupyter import get_cell_output, run_notebook
from .resource import resource_limit

log = logging.getLogger(__name__)

__all__ = [
    "DATA",
    "HIST_DF",
    "TS_DF",
    "add_random_model_data",
    "add_test_data",
    "get_cell_output",
    "make_dantzig",
    "models",
    "populate_test_platform",
    "random_model_data",
    "random_ts_data",
    "resource_limit",
    "run_notebook",
]


# Pytest hooks


def pytest_sessionstart(session):
    """Unset any configuration read from the user's directory."""
    from ixmp.backend import jdbc

    ixmp_config.clear()
    # Further clear an automatic reference to the user's home directory. See fixture
    # tmp_env below.
    ixmp_config.values["platform"]["local"].pop("path")

    # Disable slow, aggressive garbage collection
    jdbc._GC_AGGRESSIVE = False


def pytest_report_header(config, startdir):
    """Add the ixmp configuration to the pytest report header."""
    return f"ixmp config: {repr(ixmp_config.values)}"


# Session-scoped fixtures


@pytest.fixture(scope="session")
def ixmp_cli(tmp_env):
    """A CliRunner object that invokes the ixmp command-line interface."""

    class Runner(CliRunner):
        def invoke(self, *args, **kwargs):
            return super().invoke(cli.main, *args, env=tmp_env, **kwargs)

    yield Runner()


@pytest.fixture(scope="module")
def mp(test_mp):
    """A :class:`.Platform` containing test data.

    This fixture is **module** -scoped, and is used in :mod:`.test_platform`,
    :mod:`.test_timeseries`, and :mod:`.test_scenario`. :mod:`.test_meta` overrides this
    with a **function** -scoped fixture; see there for more details.
    """
    populate_test_platform(test_mp)
    yield test_mp


@pytest.fixture(scope="session")
def test_data_path():
    """Path to the directory containing test data."""
    return Path(__file__).parents[1].joinpath("tests", "data")


@pytest.fixture(scope="module")
def test_mp(request, tmp_env, test_data_path):
    """An empty :class:`Platform` connected to a temporary, in-memory database.

    This fixture has **module** scope: the same Platform is reused for all tests in a
    module.
    """
    yield from _platform_fixture(request, tmp_env, test_data_path)


@pytest.fixture(scope="session")
def tmp_env(tmp_path_factory):
    """Return the os.environ dict with the IXMP_DATA variable set.

    IXMP_DATA will point to a temporary directory that is unique to the
    test session. ixmp configuration (i.e. the 'config.json' file) can be
    written and read in this directory without modifying the current user's
    configuration.
    """
    base_temp = tmp_path_factory.getbasetemp()
    os.environ["IXMP_DATA"] = str(base_temp)

    # Set the path for the default/local platform in the test directory
    localdb = base_temp / "localdb" / "default"
    ixmp_config.values["platform"]["local"]["path"] = localdb

    # Save for other processes
    ixmp_config.save()

    yield os.environ


@pytest.fixture(scope="session")
def tutorial_path():
    """Path to the directory containing the tutorials."""
    return Path(__file__).parents[2].joinpath("tutorial")


@pytest.fixture(scope="session")
def ureg():
    """Application-wide units registry."""
    registry = pint.get_application_registry()

    # Used by .compat.ixmp, .compat.pyam
    registry.define("USD = [USD]")
    registry.define("case = [case]")

    yield registry


# Function-scoped fixtures


@pytest.fixture(scope="function")
def protect_pint_app_registry():
    """Protect pint's application registry.

    Use this fixture on tests which invoke code that calls
    :meth:`pint.set_application_registry`. It ensures that the environment for
    other tests is not altered.
    """
    import pint

    # Use deepcopy() in case the wrapped code modifies the application
    # registry without swapping out the UnitRegistry instance for a different
    # one
    saved = deepcopy(pint.get_application_registry())
    yield
    pint.set_application_registry(saved)


@pytest.fixture(scope="function")
def protect_rename_dims():
    """Protect :data:`RENAME_DIMS`.

    Use this fixture on tests which invoke code that imports :mod:`message_ix`, e.g.
    :func:`show_versions`. Importing :mod:`message_ix` has the side effect of adding
    values to :data:`RENAME_DIMS`. Using this fixture ensures that the environment for
    other tests is not altered.
    """
    from ixmp.reporting import RENAME_DIMS

    saved = deepcopy(RENAME_DIMS)  # Probably just copy() is sufficient
    yield
    RENAME_DIMS.clear()
    RENAME_DIMS.update(saved)


@pytest.fixture(scope="function")
def test_mp_f(request, tmp_env, test_data_path):
    """An empty :class:`Platform` connected to a temporary, in-memory database.

    This fixture has **function** scope: the same Platform is reused for one test
    function.

    See also
    --------
    test_mp
    """
    yield from _platform_fixture(request, tmp_env, test_data_path)


# Assertions


@contextmanager
def assert_logs(caplog, message_or_messages=None, at_level=None):
    """Assert that *message_or_messages* appear in logs.

    Use assert_logs as a context manager for a statement that is expected to trigger
    certain log messages. assert_logs checks that these messages are generated.

    Example
    -------

    def test_foo(caplog):
        with assert_logs(caplog, 'a message'):
            logging.getLogger(__name__).info('this is a message!')

    Parameters
    ----------
    caplog : object
        The pytest caplog fixture.
    message_or_messages : str or list of str
        String(s) that must appear in log messages.
    at_level : int, optional
        Messages must appear on 'ixmp' or a sub-logger with at least this level.
    """
    # Wrap a string in a list
    expected = (
        [message_or_messages]
        if isinstance(message_or_messages, str)
        else message_or_messages
    )

    # Record the number of records prior to the managed block
    first = len(caplog.records)

    if at_level is not None:
        # Use the pytest caplog fixture's built-in context manager to
        # temporarily set the level of the 'ixmp' logger
        ctx = caplog.at_level(at_level, logger="ixmp")
    else:
        # Python 3.6 compatibility: use suppress for nullcontext
        nullcontext = getattr(contextlib, "nullcontext", contextlib.suppress)
        # ctx does nothing
        ctx = nullcontext()

    try:
        with ctx:
            yield  # Nothing provided to the managed block
    finally:
        # List of bool indicating whether each of `expected` was found
        found = [any(e in msg for msg in caplog.messages[first:]) for e in expected]

        if not all(found):
            # Format a description of the missing messages
            lines = chain(
                ["Did not log:"],
                [f"    {repr(msg)}" for i, msg in enumerate(expected) if not found[i]],
                ["among:"],
                ["    []"]
                if len(caplog.records) == first
                else [f"    {repr(msg)}" for msg in caplog.messages[first:]],
            )
            pytest.fail("\n".join(lines))


# Utility functions


def bool_param_id(name):
    """Parameter ID callback for :meth:`pytest.mark.parametrize`.

    This formats a boolean value as 'name0' (False) or 'name1' (True) for
    easier selection with e.g. ``pytest -k 'name0'``.
    """
    return lambda value: "{}{}".format(name, int(value))


def create_test_platform(tmp_path, data_path, name, **properties):
    """Create a Platform for testing using specimen files '*name*.*'.

    Any of the following files from *data_path* are copied to *tmp_path*:

    - *name*.lobs, *name*.script, i.e. the contents of a :class:`.JDBCBackend`
      HyperSQL database.
    - *name*.properties.

    The contents of *name*.properties (if it exists) are formatted using the
    *properties* keyword arguments.

    Returns
    -------
    pathlib.Path
        the path to the .properties file, if any, else the .lobs file without
        suffix.
    """
    # Copy files
    any_files = False
    for suffix in ".lobs", ".properties", ".script":
        src = (data_path / name).with_suffix(suffix)
        dst = (tmp_path / name).with_suffix(suffix)
        try:
            shutil.copyfile(str(src), str(dst))
        except FileNotFoundError:
            pass
        else:
            any_files = True

    if not any_files:
        raise ValueError(f"no files for test platform {repr(name)}")

    # Create properties file
    props_file = (tmp_path / name).with_suffix(".properties")

    try:
        props = props_file.read_text()
    except FileNotFoundError:
        # No properties file; return the stub
        return tmp_path / name
    else:
        props = props.format(db_path=str(tmp_path / name), **properties)
        props_file.write_text(props)
        return props_file


# Private utilities


def _platform_fixture(request, tmp_env, test_data_path):
    """Helper for :func:`test_mp` and other fixtures."""
    # Long, unique name for the platform.
    # Remove '/' so that the name can be used in URL tests.
    platform_name = request.node.nodeid.replace("/", " ")

    # Add a platform
    ixmp_config.add_platform(
        platform_name, "jdbc", "hsqldb", url=f"jdbc:hsqldb:mem:{platform_name}"
    )

    # Launch Platform
    mp = Platform(name=platform_name)
    yield mp

    # Teardown: don't show log messages when destroying the platform, even if
    # the test using the fixture modified the log level
    mp._backend.set_log_level(logging.CRITICAL)
    del mp

    # Remove from config
    ixmp_config.remove_platform(platform_name)



from collections import ChainMap
from pathlib import Path

import pandas as pd

from ixmp.utils import maybe_check_out, maybe_commit, update_par

from .gams import GAMSModel

ITEMS = {
    # Plants
    "i": dict(ix_type="set"),
    # Markets
    "j": dict(ix_type="set"),
    # Capacity of plant i in cases
    "a": dict(ix_type="par", idx_sets=["i"]),
    # Demand at market j in cases
    "b": dict(ix_type="par", idx_sets=["j"]),
    # Distance between plant i and market j
    "d": dict(ix_type="par", idx_sets=["i", "j"]),
    # Transport cost per case per 1000 miles
    "f": dict(ix_type="par", idx_sets=None),
    # Decision variables and equations
    "x": dict(ix_type="var", idx_sets=["i", "j"]),
    "z": dict(ix_type="var", idx_sets=None),
    "cost": dict(ix_type="equ", idx_sets=None),
    "demand": dict(ix_type="equ", idx_sets=["j"]),
    "supply": dict(ix_type="equ", idx_sets=["i"]),
}

DATA = {
    "i": ["seattle", "san-diego"],
    "j": ["new-york", "chicago", "topeka"],
    "a": pd.DataFrame(
        [
            ["seattle", 350, "cases"],
            ["san-diego", 600, "cases"],
        ],
        columns="i value unit".split(),
    ),
    "b": pd.DataFrame(
        [
            ["new-york", 325, "cases"],
            ["chicago", 300, "cases"],
            ["topeka", 275, "cases"],
        ],
        columns="j value unit".split(),
    ),
    "d": pd.DataFrame(
        [
            ["seattle", "new-york", 2.5, "km"],
            ["seattle", "chicago", 1.7, "km"],
            ["seattle", "topeka", 1.8, "km"],
            ["san-diego", "new-york", 2.5, "km"],
            ["san-diego", "chicago", 1.8, "km"],
            ["san-diego", "topeka", 1.4, "km"],
        ],
        columns="i j value unit".split(),
    ),
    "f": (90.0, "USD/km"),
}


class DantzigModel(GAMSModel):
    """Dantzig's cannery/transport problem as a :class:`GAMSModel`.

    Provided for testing :mod:`ixmp` code.
    """

    name = "dantzig"

    defaults = ChainMap(
        {
            # Override keys from GAMSModel
            "model_file": Path(__file__).with_name("dantzig.gms"),
        },
        GAMSModel.defaults,
    )

    @classmethod
    def initialize(cls, scenario, with_data=False):
        """Initialize the problem.

        If *with_data* is :obj:`True` (default: :obj:`False`), the set and parameter
        values from the original problem are also populated. Otherwise, the sets and
        parameters are left empty.
        """
        # Initialize the ixmp items
        cls.initialize_items(scenario, ITEMS)

        if not with_data:
            return

        checkout = maybe_check_out(scenario)

        # Add set elements
        scenario.add_set("i", DATA["i"])
        scenario.add_set("j", DATA["j"])

        # Add parameter values
        update_par(scenario, "a", DATA["a"])
        update_par(scenario, "b", DATA["b"])
        update_par(scenario, "d", DATA["d"])

        # TODO avoid overwriting the existing value
        scenario.change_scalar("f", *DATA["f"])

        maybe_commit(scenario, checkout, f"{cls.__name__}.initialize")



import logging
import os
import re
from abc import ABC, abstractmethod

from ixmp.utils import maybe_check_out, maybe_commit

log = logging.getLogger(__name__)


class ModelError(Exception):
    """Error in model code, e.g. :meth:`.Model.run`."""


class Model(ABC):
    #: Name of the model.
    name = "base"

    @abstractmethod
    def __init__(self, name, **kwargs):
        """Constructor.

        Parameters
        ----------
        kwargs :
            Model options, passed directly from :meth:`.Scenario.solve`.

            Model subclasses MUST document acceptable option values.
        """

    @classmethod
    def clean_path(cls, value: str) -> str:
        """Subtitute invalid characters in `value` with "_"."""
        chars = r'<>"/\|?*' + (":" if os.name == "nt" else "")
        return re.sub("[{}]+".format(re.escape(chars)), "_", value)

    @classmethod
    def initialize(cls, scenario):
        """Set up *scenario* with required items.

        Implementations of :meth:`initialize`:

        - **may** add sets, set elements, and/or parameter values.
        - **may** accept any number of keyword arguments to control behaviour.
        - **must not** modify existing parameter data in *scenario*, either by
          deleting or overwriting values.

        Parameters
        ----------
        scenario : .Scenario
            Object to initialize.

        See also
        --------
        initialize_items
        """
        log.debug(f"No initialization for {repr(scenario.scheme)}-scheme Scenario")

    @classmethod
    def initialize_items(cls, scenario, items):
        """Helper for :meth:`initialize`.

        All of the `items` are added to `scenario`. Existing items are not
        modified. Errors are logged if the description in `items` conflicts
        with the index set(s) and/or index name(s) of existing items.

        initialize_items may perform one commit. `scenario` is in the same
        state (checked in, or checked out) after initialize_items is complete.

        Parameters
        ----------
        scenario : .Scenario
            Object to initialize.
        items : dict of (str -> dict)
            Each key is the name of an ixmp item (set, parameter, equation, or
            variable) to initialize. Each dict **must** have the key 'ix_type';
            one of 'set', 'par', 'equ', or 'var'; any other entries are keyword
            arguments to the methods :meth:`.init_set` etc.

        Raises
        ------
        ValueError
            if `scenario` has a solution, i.e. :meth:`~.Scenario.has_solution`
            is :obj:`True`.

        See also
        --------
        .init_equ
        .init_par
        .init_set
        .init_var
        """
        # Don't know if the Scenario is checked out
        checkout = None

        # Lists of items in the Scenario
        existing_items = dict()

        # Lists of items initialized
        items_initialized = []

        for name, item_info in items.items():
            # Copy so that pop() below does not modify *items*
            item_info = item_info.copy()

            # Check that the item exists
            ix_type = item_info.pop("ix_type")

            if ix_type not in existing_items:
                # Store a list of items of *ix_type*
                method = getattr(scenario, f"{ix_type}_list")
                existing_items[ix_type] = method()

            # Item must be initialized if it does not exist

            if name in existing_items[ix_type]:
                # Item exists; check its index sets and names
                for key, values in item_info.items():
                    values = values or []
                    existing = getattr(scenario, key)(name)
                    if existing != list(values):
                        # The existing index sets or names do not match
                        log.warning(
                            f"Existing index {key.split('_')[-1]} of "
                            f"{repr(name)} {repr(existing)} do not match "
                            f"{repr(values)}"
                        )

                # Skip; can't do anything to existing items
                continue

            # Item doesn't exist and must be initialized

            # Possibly check out the Scenario
            try:
                checkout = maybe_check_out(scenario, checkout)
            except ValueError as exc:  # pragma: no cover
                # The Scenario has a solution. This indicates an inconsistent
                # situation: the Scenario lacks the item *name*, but somehow it
                # was successfully solved without it, and the solution stored.
                # Can't proceed further.
                log.error(str(exc))
                return

            # Get the appropriate method, e.g. init_set, and add the item
            log.info(f"Initialize {ix_type} {repr(name)} as {item_info}")
            getattr(scenario, f"init_{ix_type}")(name=name, **item_info)

            # Record
            items_initialized.append(name)

        maybe_commit(
            scenario, len(items_initialized), f"{cls.__name__}.initialize_items"
        )

        if len(items_initialized) and not checkout:
            # Scenario was originally in a checked out state; restore
            maybe_check_out(scenario)

    @abstractmethod
    def run(self, scenario):
        """Execute the model.

        Parameters
        ----------
        scenario : .Scenario
            Scenario object to solve by running the Model.
        """



import genno.config
from genno import (
    ComputationError,
    Computer,
    Key,
    KeyExistsError,
    MissingKeyError,
    Quantity,
    configure,
)

from .reporter import Reporter
from .util import RENAME_DIMS

__all__ = [
    # ixmp-specific
    "RENAME_DIMS",
    "Reporter",
    # Re-exports from genno
    "ComputationError",
    "Key",
    "KeyExistsError",
    "MissingKeyError",
    "Quantity",
    "configure",
]


@genno.config.handles("filters", iterate=False)
def filters(c: Computer, filters: dict):
    """Handle the entire ``filters:`` config section."""
    # Ensure a filters dict exists
    c.graph["config"].setdefault("filters", dict())

    if len(filters):
        # Store the new filters
        c.graph["config"]["filters"].update(filters)
    else:
        # Empty dictionary → clear all
        c.graph["config"]["filters"].clear()

    # Clear filters for specific dimensions only
    for key, value in filters.items():
        if value is None:
            c.graph["config"]["filters"].pop(key, None)


@genno.config.handles("rename_dims", iterate=False)
def rename_dims(c: Computer, info: dict):
    """Handle the entire ``rename_dims:`` config section."""
    RENAME_DIMS.update(info)


# keep=True is different vs. genno.config
@genno.config.handles("units", iterate=False, discard=False)
def units(c: Computer, info: dict):
    """Handle the entire ``units:`` config section."""
    genno.config.units(c, info)



# Methods are in alphabetical order
from itertools import product
from math import ceil
from typing import Any, List

import numpy as np
import pandas as pd
import pint
import xarray as xr

from ixmp.core import IAMC_IDX, Platform, Scenario, TimeSeries
from ixmp.reporting import Quantity

#: Common (model name, scenario name) pairs for testing.
SCEN = {
    "dantzig": dict(model="canning problem", scenario="standard"),
    "h2g2": dict(model="Douglas Adams", scenario="Hitchhiker"),
}
models = SCEN

_MS: List[Any] = [models["dantzig"]["model"], models["dantzig"]["scenario"]]
HIST_DF = pd.DataFrame(
    [_MS + ["DantzigLand", "GDP", "USD", 850.0, 900.0, 950.0]],
    columns=IAMC_IDX + [2000, 2005, 2010],
)
INP_DF = pd.DataFrame(
    [_MS + ["DantzigLand", "Demand", "cases", 850.0, 900.0]],
    columns=IAMC_IDX + [2000, 2005],
)
TS_DF = (
    pd.concat([HIST_DF, INP_DF], sort=False)
    .sort_values(by="variable")
    .reset_index(drop=True)
)

# Test data used by test_scenario.py and test_timeseries.py.
#
# NB the columns are in a specific order; model and scenario come last in the data
#    returned by ixmp.
# TODO fix this; model and scenario should come first, matching the IAMC order.
DATA = {
    0: pd.DataFrame.from_dict(
        dict(
            region="World",
            variable="Testing",
            unit="???",
            year=[2010, 2020],
            value=[23.7, 23.8],
            model="model name",
            scenario="scenario name",
        )
    ),
    2010: pd.DataFrame.from_dict(
        {
            "region": ["World"],
            "variable": ["Testing"],
            "unit": ["???"],
            "2010": [23.7],
            "2020": [23.8],
        }
    ),
    2030: pd.DataFrame.from_dict(
        dict(
            region="World",
            variable=["Testing", "Testing", "Testing2"],
            unit="???",
            year=[2020, 2030, 2030],
            value=[24.8, 24.9, 25.1],
            model="model name",
            scenario="scenario name",
        )
    ),
    2050: pd.DataFrame.from_dict(
        dict(
            region="World",
            variable="Testing",
            unit="???",
            year=[2000, 2010, 2020, 2030, 2040, 2050],
            value=[21.7, 22.7, 23.7, 24.7, 25.7, 25.8],
            model="model name",
            scenario="scenario name",
        )
    ),
    # NB the columns for geodata methods are inconsistent with those for time-series
    #    data
    "geo": pd.DataFrame.from_dict(
        dict(
            region="World",
            variable="var1",
            subannual="Year",
            year=[2000, 2010, 2020],
            value=["test", "more-test", "2020-test"],
            unit="score",
            meta=0,
        )
    ),
    "timeseries": pd.DataFrame.from_dict(
        dict(
            region="World",
            variable="Testing",
            unit="???",
            year=[2010, 2020],
            value=[23.7, 23.8],
        )
    ),
}


def add_random_model_data(scenario, length):
    """Add a set and parameter with given *length* to *scenario*.

    The set is named 'random_set'. The parameter is named 'random_par', and has two
    dimensions indexed by 'random_set'.
    """
    set_data, par_data = random_model_data(length)
    scenario.init_set("random_set")
    scenario.add_set("random_set", set_data)
    scenario.init_par(
        "random_par",
        idx_sets=["random_set", "random_set"],
        idx_names=["random_set0", "random_set1"],
    )
    scenario.add_par("random_par", par_data)
    return len(par_data)


def add_test_data(scen: Scenario):
    # New sets
    t_foo = ["foo{}".format(i) for i in (1, 2, 3)]
    t_bar = ["bar{}".format(i) for i in (4, 5, 6)]
    t = t_foo + t_bar
    y = list(map(str, range(2000, 2051, 10)))

    # Add to scenario
    scen.init_set("t")
    scen.add_set("t", t)
    scen.init_set("y")
    scen.add_set("y", y)

    # Data
    ureg = pint.get_application_registry()
    x = Quantity(
        xr.DataArray(np.random.rand(len(t), len(y)), coords=[("t", t), ("y", y)]),
        units=ureg.kg,
    )

    # As a pd.DataFrame with units
    x_df = x.to_series().rename("value").reset_index()
    x_df["unit"] = "kg"

    scen.init_par("x", ["t", "y"])
    scen.add_par("x", x_df)

    return t, t_foo, t_bar, x


def make_dantzig(mp: Platform, solve: bool = False, quiet: bool = False) -> Scenario:
    """Return :class:`ixmp.Scenario` of Dantzig's canning/transport problem.

    Parameters
    ----------
    mp : .Platform
        Platform on which to create the scenario.
    solve : bool, optional
        If :obj:`True`. then solve the scenario before returning. Default :obj:`False`.
    quiet : bool, optional
        If :obj:`True`, suppress console output when solving.

    Returns
    -------
    .Scenario

    See also
    --------
    .DantzigModel
    """
    # add custom units and region for timeseries data
    try:
        mp.add_unit("USD/km")
    except Exception:
        # Unit already exists. Pending bugfix from zikolach
        pass
    mp.add_region("DantzigLand", "country")

    # Initialize a new Scenario, and use the DantzigModel class' initialize()
    # method to populate it
    annot = "Dantzig's transportation problem for illustration and testing"
    scen = Scenario(
        mp,
        **models["dantzig"],  # type: ignore [arg-type]
        version="new",
        annotation=annot,
        scheme="dantzig",
        with_data=True,
    )

    # commit the scenario
    scen.commit("Import Dantzig's transport problem for testing.")

    # set this new scenario as the default version for the model/scenario name
    scen.set_as_default()

    if solve:
        # Solve the model using the GAMS code provided in the `tests` folder
        scen.solve(model="dantzig", case="transport_standard", quiet=quiet)

    # add timeseries data for testing `clone(keep_solution=False)`
    # and `remove_solution()`
    scen.check_out(timeseries_only=True)
    scen.add_timeseries(HIST_DF, meta=True)
    scen.add_timeseries(INP_DF)
    scen.commit("Import Dantzig's transport problem for testing.")

    return scen


def populate_test_platform(platform):
    """Populate `platform` with data for testing.

    Many of the tests in :mod:`ixmp.tests.core` depend on this set of data.

    The data consist of:

    - 3 versions of the Dantzig cannery/transport Scenario.

      - Version 2 is the default.
      - All have :obj:`HIST_DF` and :obj:`TS_DF` as time-series data.

    - 1 version of a TimeSeries with model name 'Douglas Adams' and scenario
      name 'Hitchhiker', containing 2 values.
    """
    s1 = make_dantzig(platform, solve=True, quiet=True)

    s2 = s1.clone()
    s2.set_as_default()

    s2.clone()

    s4 = TimeSeries(platform, **models["h2g2"], version="new")
    s4.add_timeseries(
        pd.DataFrame.from_dict(
            dict(
                region="World",
                variable="Testing",
                unit="???",
                year=[2010, 2020],
                value=[23.7, 23.8],
            )
        )
    )
    s4.commit("")
    s4.set_as_default()


def random_model_data(length):
    """Random (set, parameter) data with at least *length* elements.

    See also
    --------
    add_random_model_data
    """
    # Dimension size
    dim_len = ceil(length ** 0.5)
    set_data = list(str(i) for i in range(dim_len))

    # Revised length, possibly slightly higher than original
    length = dim_len ** 2

    par_data = pd.concat(
        [
            pd.DataFrame.from_dict(
                dict(region="World", value=np.random.rand(length), unit="GWa")
            ),
            pd.DataFrame(
                data=product(set_data, set_data), columns=["random_set0", "random_set1"]
            ),
        ],
        axis=1,
    )

    return set_data, par_data


def random_ts_data(length):
    """A :class:`pandas.DataFrame` of time series data with *length* rows.

    Suitable for passage to :meth:`TimeSeries.add_timeseries`.
    """
    return pd.DataFrame.from_dict(
        dict(
            region="World",
            variable=[f"foo|{i}" for i in range(int(length))],
            year=2020,
            value=np.random.rand(int(length)),
            unit="GWa",
        )
    )



import logging
from collections import deque

import pandas as pd

from ixmp.utils import as_str_list, maybe_check_out, maybe_commit

from . import ItemType

log = logging.getLogger(__name__)

#: Maximum number of rows supported by the Excel file format. See
#: :meth:`.to_excel` and :ref:`excel-data-format`.
EXCEL_MAX_ROWS = 1048576


def ts_read_file(ts, path, firstyear=None, lastyear=None):
    """Read data from a CSV or Microsoft Excel file at *path* into *ts*.

    See also
    --------
    TimeSeries.add_timeseries
    TimeSeries.read_file
    """

    if path.suffix == ".csv":
        df = pd.read_csv(path)
    elif path.suffix == ".xlsx":
        df = pd.read_excel(path)

    ts.check_out(timeseries_only=True)
    ts.add_timeseries(df, year_lim=(firstyear, lastyear))

    msg = f"adding timeseries data from {path}"
    if firstyear:
        msg += f" from {firstyear}"
    if lastyear:
        msg += f" until {lastyear}"
    ts.commit(msg)


def s_write_excel(be, s, path, item_type, filters=None, max_row=None):
    """Write *s* to a Microsoft Excel file at *path*.

    See also
    --------
    Scenario.to_excel
    """
    # Default: empty dict
    filters = filters or dict()

    # Types of items to write
    ix_types = ["set", "par"]
    if ItemType.VAR in item_type:
        ix_types.append("var")
    if ItemType.EQU in item_type:
        ix_types.append("equ")

    # item name -> ixmp type
    name_type = {}
    for ix_type in ix_types:
        names = sorted(be.list_items(s, ix_type))
        name_type.update({n: ix_type for n in names})

    # Open file
    writer = pd.ExcelWriter(path)

    omitted = set()
    empty_sets = []

    # Don't allow the user to set a value that will truncate data
    max_row = min(int(max_row or EXCEL_MAX_ROWS), EXCEL_MAX_ROWS)

    for name, ix_type in name_type.items():
        if ix_type == "par":
            # Use only the filters corresponding to dimensions of this item
            item_filters = {
                k: v for k, v in filters.items() if k in be.item_index(s, name, "names")
            }
        else:
            item_filters = None

        # Extract data: dict, pd.Series, or pd.DataFrame
        data = be.item_get_elements(s, ix_type, name, item_filters)

        if isinstance(data, dict):
            # Scalar equ/par/var: series with index like 'value', 'unit'.
            # Convert to DataFrame with 1 row.
            data = pd.Series(data, name=name).to_frame().transpose()
        elif isinstance(data, pd.Series):
            # Index set: use own name as the header
            data.name = name

        if data.empty:
            if ix_type != "set":
                # Don't write empty equ/par/var
                omitted.add(name)
            else:
                # Write empty sets later
                empty_sets.append((name, data))
            continue

        # Write data in multiple sheets

        # List of break points, plus the final row
        splits = list(range(0, len(data), max_row)) + [len(data)]
        # Pairs of row numbers, e.g. (0, 100), (100, 200), ...
        first_last = zip(splits, splits[1:])

        for sheet_num, (first_row, last_row) in enumerate(first_last, start=1):
            # Format the sheet name, possibly with a suffix
            sheet_name = name + (f"({sheet_num})" if sheet_num > 1 else "")

            # Subset the data (only on rows, if a DataFrame) and write
            data.iloc[first_row:last_row].to_excel(writer, sheet_name, index=False)

    # Discard entries that were not written
    for name in omitted:
        name_type.pop(name)

    # Write the name -> type map
    pd.Series(name_type, name="ix_type").rename_axis(
        index="item"
    ).reset_index().to_excel(writer, sheet_name="ix_type_mapping", index=False)

    # Write empty sets last
    for name, data in empty_sets:
        data.to_excel(writer, sheet_name=name, index=False)

    writer.save()


def maybe_init_item(scenario, ix_type, name, new_idx, path):
    """Call :meth:`~.init_set`, :meth:`.init_par`, etc. if possible.

    Logs an intelligible warning and then raises ValueError in two cases:

    - the *new_idx* is ambiguous, e.g. containing index names that cannot be
      used to infer index sets, or
    - an existing item has index names that are different from *new_idx*.

    """
    try:
        # [] and None are equivalent; convert to be consistent
        existing_names = scenario.idx_names(name) or None
    except KeyError:
        # Item does not exist

        # Check for ambiguous index names
        ambiguous_idx = sorted(set(new_idx or []) - set(scenario.set_list()))
        if len(ambiguous_idx):
            log.warning(
                f"Cannot read {ix_type} {repr(name)}: index set(s) cannot be "
                f"inferred for name(s) {ambiguous_idx}"
            )
            raise ValueError from None

        # Initialize
        getattr(scenario, f"init_{ix_type}")(name, new_idx)
    else:
        # Item exists; check that is has the same index names

        # [] and None are equivalent; convert to be consistent
        if isinstance(new_idx, list) and new_idx == []:
            new_idx = None

        if existing_names != new_idx:
            log.warning(
                f"Existing {ix_type} {repr(name)} has index names(s) "
                f" {existing_names} != {new_idx} in {path.name}"
            )
            raise ValueError from None


def s_read_excel(be, s, path, add_units=False, init_items=False, commit_steps=False):
    """Read data from a Microsoft Excel file at *path* into *s*.

    See also
    --------
    Scenario.read_excel
    """
    log.info(f"Read data from {path}")

    # Get item name -> ixmp type mapping as a pd.Series
    xf = pd.ExcelFile(path, engine="openpyxl")
    name_type = xf.parse("ix_type_mapping", index_col="item")["ix_type"]

    # Queue of (set name, data) to add
    sets_to_add = deque((n, None) for n in name_type.index[name_type == "set"])

    def parse_item_sheets(name):
        """Read data for item *name*, possibly across multiple sheets."""
        dfs = [xf.parse(name)]

        # Collect data from repeated sheets due to max_row limit
        for x in filter(lambda n: n.startswith(name + "("), xf.sheet_names):
            dfs.append(xf.parse(x))

        # Concatenate once and return
        return pd.concat(dfs, axis=0)

    # Add sets in two passes:
    # 1. Index sets, required to initialize other sets.
    # 2. Sets indexed by others.
    while True:
        try:
            # Get an item from the queue
            name, data = sets_to_add.popleft()
        except IndexError:
            break  # Finished

        log.info(name)

        first_pass = data is None
        if first_pass:
            # Read data
            data = parse_item_sheets(name)

        # Determine index set(s) for this set
        idx_sets = data.columns.to_list()
        if len(idx_sets) == 1:
            if idx_sets == [0]:  # pragma: no cover
                # Old-style export with uninformative '0' as a column header;
                # assume it is an index set
                log.warning(f"Add {name} with header '0' as index set")
                idx_sets = None
            elif idx_sets == [name]:
                # Set's own name as column header -> an index set
                idx_sets = None
            else:
                pass  # 1-D set indexed by another set

        if first_pass and idx_sets is not None:
            # Indexed set; append to the queue to process later
            sets_to_add.append((name, data))
            continue

        # At this point: either an index set, or second pass when all index
        # sets have been init'd and populated
        if init_items:
            try:
                maybe_init_item(s, "set", name, idx_sets, path)
            except ValueError:
                continue  # Ambiguous or conflicting; skip this set

        # Convert data as expected by add_set
        if len(data.columns) == 1:
            # Convert data frame into 1-D vector
            data = data.iloc[:, 0].values

            if idx_sets is not None:
                # Indexed set must be input as list of list of str
                data = list(map(as_str_list, data))

        try:
            s.add_set(name, data)
        except KeyError:
            raise ValueError(f"no set {repr(name)}; try init_items=True")

    maybe_commit(s, commit_steps, f"Loaded sets from {path}")

    # List of existing units for reference
    units = set(be.get_units())

    # Add equ/par/var data
    for name, ix_type in name_type[name_type != "set"].items():
        if ix_type in ("equ", "var"):
            log.info(f"Cannot read {ix_type} {repr(name)}")
            continue

        # Only parameters beyond this point

        df = parse_item_sheets(name)

        maybe_check_out(s)

        if add_units:
            # New units appearing in this parameter
            to_add = set(df["unit"].unique()) - units

            for unit in to_add:
                log.info(f"Add missing unit: {unit}")
                # FIXME cannot use the comment f'Loaded from {path}' here; too
                #       long for JDBCBackend
                be.set_unit(unit, "Loaded from file")

            # Update the reference set to avoid re-adding these units
            units |= to_add

        # NB if equ/var were imported, also need to filter 'lvl', 'mrg' here
        idx_sets = list(filter(lambda v: v not in ("value", "unit"), df.columns))

        if init_items:
            try:
                # Same as init_scalar if idx_sets == []
                maybe_init_item(s, ix_type, name, idx_sets, path)
            except ValueError:
                continue  # Ambiguous or conflicting; skip this parameter

        if not len(idx_sets):
            # No index sets -> scalar parameter; must supply empty 'key' column
            # for add_par()
            df["key"] = None

        s.add_par(name, df)

        # Commit after every parameter
        maybe_commit(s, commit_steps, f"Loaded {ix_type} {repr(name)} from {path}")

    maybe_commit(s, not commit_steps, f"Import from {path}")



from pathlib import Path

import pandas as pd
import pytest
from click.exceptions import BadParameter, UsageError
from pandas.testing import assert_frame_equal

import ixmp
from ixmp.cli import VersionType
from ixmp.testing import models, populate_test_platform


def test_versiontype():
    vt = VersionType()
    # str converts to int
    assert vt.convert("1", None, None) == 1
    assert vt.convert("-1", None, None) == -1

    # str 'new' is passes through
    assert vt.convert("new", None, None) == "new"

    # int passes through
    assert vt.convert(1, None, None) == 1

    with pytest.raises(BadParameter, match="'xx' must be an integer or 'new'"):
        vt.convert("xx", None, None)


def test_main(ixmp_cli, test_mp, tmp_path):
    # Name of a temporary file that doesn't exist
    tmp_path /= "temp.properties"

    # Giving --dbprops and a nonexistent file is an invalid argument
    cmd = [
        "--platform",
        "pname",
        "--dbprops",
        str(tmp_path),
        "platform",
        "list",  # Doesn't get executed; fails in cli.main()
    ]
    result = ixmp_cli.invoke(cmd)
    # Check against click's default exit code for the exception
    assert result.exit_code == UsageError.exit_code

    # Create the file
    tmp_path.write_text("")

    # Giving both --platform and --dbprops is bad option usage
    result = ixmp_cli.invoke(cmd)
    assert result.exit_code == UsageError.exit_code

    # --dbprops alone causes backend='jdbc' to be inferred (but an error
    # because temp.properties is empty)
    result = ixmp_cli.invoke(cmd[2:])
    assert "Config file contains no database URL" in result.exception.args[0]

    # --url argument can be given
    cmd = [
        "--url",
        "ixmp://{}/Douglas Adams/Hitchhiker".format(test_mp.name),
        "platform",
        "list",
    ]
    result = ixmp_cli.invoke(cmd)
    assert result.exit_code == 0

    # --url and other Platform/Scenario specifiers are bad option usage
    result = ixmp_cli.invoke(["--platform", "foo"] + cmd)
    assert result.exit_code == UsageError.exit_code


def test_config(ixmp_cli):
    # ixmp has no string keys by default, so we insert a fake one
    ixmp.config.register("test key", str)
    ixmp.config.values["test key"] = "foo"

    # show() works
    assert ixmp_cli.invoke(["config", "show"]).output.startswith("Configuration path: ")

    # get() works
    assert ixmp_cli.invoke(["config", "get", "test key"]).output == "foo\n"

    # set() changes the value
    result = ixmp_cli.invoke(["config", "set", "test key", "bar"])
    assert result.exit_code == 0
    assert ixmp_cli.invoke(["config", "get", "test key"]).output == "bar\n"

    # get() with a value is an invalid call
    result = ixmp_cli.invoke(["config", "get", "test key", "BADVALUE"])
    assert result.exit_code != 0


def test_list(ixmp_cli, test_mp):
    cmd = ["list", "--match", "foo"]

    # 'list' without specifying a platform/scenario is a UsageError
    result = ixmp_cli.invoke(cmd)
    assert result.exit_code == UsageError.exit_code

    # CLI works; nothing returned with a --match option that matches nothing
    result = ixmp_cli.invoke(["--platform", test_mp.name] + cmd)
    assert result.exit_code == 0, (result.exception, result.output)
    assert (
        result.output
        == """
0 model name(s)
0 scenario name(s)
0 (model, scenario) combination(s)
0 total scenarios
"""
    )


def test_platform(ixmp_cli, tmp_path):
    """Test 'platform' command."""

    def call(*args, exit_0=True):
        result = ixmp_cli.invoke(["platform"] + list(map(str, args)))
        assert not exit_0 or result.exit_code == 0
        return result

    # The default platform is 'local'
    r = call("list")
    assert "default: local\n" in r.output

    # JBDC Oracle platform can be added
    r = call("add", "p1", "jdbc", "oracle", "HOSTNAME", "USER", "PASSWORD")

    # Default platform can be changed
    r = call("add", "default", "p1")
    r = call("list")
    assert "default: p1\n" in r.output
    # Reset to avoid disturbing other tests
    call("add", "default", "local")

    # Setting the default using a non-existent platform fails
    r = call("add", "default", "nonexistent", exit_0=False)
    assert r.exit_code == 1

    # JDBC HSQLDB platform can be added with absolute path
    r = call("add", "p2", "jdbc", "hsqldb", tmp_path)
    assert ixmp.config.get_platform_info("p2")[1]["path"] == tmp_path

    # JDBC HSQLDB platform can be added with relative path
    rel = "./foo"
    r = call("add", "p3", "jdbc", "hsqldb", rel)
    assert Path(rel).resolve() == ixmp.config.get_platform_info("p3")[1]["path"]

    # Platform can be removed
    r = call("remove", "p3")
    assert r.output == "Removed platform config for 'p3'\n"

    # Non-existent platform can't be removed
    r = call("remove", "p3", exit_0=False)  # Already removed
    assert r.exit_code == 1

    # Extra args to 'remove' are invalid
    r = call("remove", "p2", "BADARG", exit_0=False)
    assert UsageError.exit_code == r.exit_code


def test_import_ts(ixmp_cli, test_mp, test_data_path):
    # Ensure the 'canning problem'/'standard' TimeSeries exists
    populate_test_platform(test_mp)

    # Invoke the CLI to import data to version 1 of the TimeSeries
    result = ixmp_cli.invoke(
        [
            "--platform",
            test_mp.name,
            "--model",
            models["dantzig"]["model"],
            "--scenario",
            models["dantzig"]["scenario"],
            "--version",
            "1",
            "import",
            "timeseries",
            "--firstyear",
            "2020",
            "--lastyear",
            "2200",
            str(test_data_path / "timeseries_canning.csv"),
        ]
    )
    assert result.exit_code == 0, result.output

    # Expected data
    exp = pd.DataFrame.from_dict(
        {
            "region": ["World"],
            "variable": ["Testing"],
            "unit": ["???"],
            "year": [2020],
            "value": [28.3],
            "model": ["canning problem"],
            "scenario": ["standard"],
        }
    )

    # The specified TimeSeries version contains the expected data
    scen = ixmp.Scenario(test_mp, **models["dantzig"], version=1)
    assert_frame_equal(scen.timeseries(variable=["Testing"]), exp)

    # The data is not present in other versions
    scen = ixmp.Scenario(test_mp, **models["dantzig"], version=2)
    assert len(scen.timeseries(variable=["Testing"])) == 0


def test_excel_io(ixmp_cli, test_mp, tmp_path):
    populate_test_platform(test_mp)
    tmp_path /= "dantzig.xlsx"

    url = (
        f"ixmp://{test_mp.name}/{models['dantzig']['model']}/"
        f"{models['dantzig']['scenario']}"
    )

    # Invoke the CLI to export data to Excel
    cmd = ["--url", url, "export", str(tmp_path)]
    result = ixmp_cli.invoke(cmd)
    assert result.exit_code == 0, result.output

    # Export with a maximum row limit per sheet
    tmp_path2 = tmp_path.with_name("dantzig2.xlsx")
    cmd = cmd[:-1] + [str(tmp_path2), "--max-row", 2]
    result = ixmp_cli.invoke(cmd)
    assert result.exit_code == 0, result.output

    # Fails without platform/scenario info
    assert ixmp_cli.invoke(cmd[2:]).exit_code == UsageError.exit_code

    # Invoke the CLI to read data from Excel
    cmd = ["--url", url, "import", "scenario", str(tmp_path)]

    # Fails without platform/scenario info
    assert ixmp_cli.invoke(cmd[2:]).exit_code == UsageError.exit_code

    # Fails without --discard-solution
    result = ixmp_cli.invoke(cmd)
    assert result.exit_code == 1
    assert "This Scenario has a solution" in result.output

    # Succeeds with --discard-solution
    cmd.insert(-1, "--discard-solution")
    result = ixmp_cli.invoke(cmd)
    assert result.exit_code == 0, result.output

    # Import into a new model name: fails without --init-items
    cmd = [
        "--url",
        f"ixmp://{test_mp.name}/foo model/bar scenario#new",
        "import",
        "scenario",
        str(tmp_path),
    ]
    result = ixmp_cli.invoke(cmd)
    assert result.exit_code == 1, result.output

    # Succeeds with --init-items
    cmd.insert(-1, "--init-items")
    result = ixmp_cli.invoke(cmd)
    assert result.exit_code == 0, result.output

    # Import from a file that has multiple sheets (due to row limit)
    cmd[-1] = str(tmp_path2)
    result = ixmp_cli.invoke(cmd)
    assert result.exit_code == 0, result.output


def test_excel_io_filters(ixmp_cli, test_mp, tmp_path):
    populate_test_platform(test_mp)
    tmp_path /= "dantzig.xlsx"

    url = (
        f"ixmp://{test_mp.name}/{models['dantzig']['model']}/"
        f"{models['dantzig']['scenario']}"
    )

    # Invoke the CLI to export data to Excel, with filters
    cmd = [
        "--url",
        url,
        "export",
        str(tmp_path),
        "--",
        "i=seattle",
    ]
    result = ixmp_cli.invoke(cmd)
    assert result.exit_code == 0, result.output

    # Import into a new model name
    url = f"ixmp://{test_mp.name}/foo model/bar scenario#new"
    cmd = [
        "--url",
        url,
        "import",
        "scenario",
        "--init-items",
        str(tmp_path),
    ]
    result = ixmp_cli.invoke(cmd)
    assert result.exit_code == 0, result.output

    # Load one of the imported parameters
    scen = ixmp.Scenario(test_mp, "foo model", "bar scenario")
    d = scen.par("d")

    # Data in (imported from) file has only filtered elements
    assert set(d["i"].unique()) == {"seattle"}
    assert len(d) == 3


def test_report(ixmp_cli):
    # 'report' without specifying a platform/scenario is a UsageError
    result = ixmp_cli.invoke(["report", "key"])
    assert result.exit_code == UsageError.exit_code


@pytest.mark.usefixtures("protect_pint_app_registry", "protect_rename_dims")
def test_show_versions(ixmp_cli):
    result = ixmp_cli.invoke(["show-versions"])
    assert result.exit_code == 0, result.output


def test_solve(ixmp_cli, test_mp):
    populate_test_platform(test_mp)
    cmd = [
        "--platform",
        test_mp.name,
        "--model",
        models["dantzig"]["model"],
        "--scenario",
        models["dantzig"]["scenario"],
        "solve",
        "--remove-solution",
    ]
    result = ixmp_cli.invoke(cmd)
    assert result.exit_code == 0, result.output

    # test failing solving without solution removal
    cmd = [
        "--platform",
        test_mp.name,
        "--model",
        models["dantzig"]["model"],
        "--scenario",
        models["dantzig"]["scenario"],
        "solve",
    ]
    result = ixmp_cli.invoke(cmd)
    assert result.exit_code == 1, result.output

    # missing scenario
    cmd = [
        "--platform",
        test_mp.name,
        "--model",
        "non-existing",
        "--scenario",
        models["dantzig"]["scenario"],
        "solve",
    ]
    result = ixmp_cli.invoke(cmd)
    assert result.exit_code == 1, result.output
    assert "Error: model='non-existing'" in result.output

    # no platform/scenario provided
    cmd = ["solve"]
    result = ixmp_cli.invoke(cmd)
    assert result.exit_code != 0, result.output
    assert "Error: give --url before command solve" in result.output



pytest_plugins = ("genno.testing", "ixmp.testing")


# Hooks


def pytest_addoption(parser):
    parser.addoption(
        "--jvm-mem-limit",
        action="store",
        default=-1,
        help=(
            "Memory limit, in MiB, for the Java Virtual Machine (JVM) started by the "
            "ixmp JDBCBackend"
        ),
    )
    parser.addoption(
        "--resource-limit",
        action="store",
        default="DATA:-1",
        help=(
            "Limit a Python resource via the ixmp.testing.resource_limit fixture. Use "
            "e.g. 'DATA:500' to limit RLIMIT_DATA to 500 MiB"
        ),
    )



import logging
from itertools import zip_longest

import pandas as pd
import pint
from genno.core.quantity import Quantity
from genno.util import parse_units

from .util import RENAME_DIMS, dims_for_qty, get_reversed_rename_dims

log = logging.getLogger(__name__)


def data_for_quantity(ix_type, name, column, scenario, config):
    """Retrieve data from *scenario*.

    Parameters
    ----------
    ix_type : 'equ' or 'par' or 'var'
        Type of the ixmp object.
    name : str
        Name of the ixmp object.
    column : 'mrg' or 'lvl' or 'value'
        Data to retrieve. 'mrg' and 'lvl' are valid only for ``ix_type='equ'``,and
        'level' otherwise.
    scenario : ixmp.Scenario
        Scenario containing data to be retrieved.
    config : dict of (str -> dict)
        The key 'filters' may contain a mapping from dimensions to iterables of allowed
        values along each dimension. The key 'units'/'apply' may contain units to apply
        to the quantity; any such units overwrite existing units, without conversion.

    Returns
    -------
    :class:`Quantity`
        Data for *name*.
    """
    log.debug(f"{name}: retrieve data")

    registry = pint.get_application_registry()

    # Only use the relevant filters
    idx_names = scenario.idx_names(name)
    filters = config.get("filters", None)
    if filters:
        # Dimensions of the object
        dims = dims_for_qty(idx_names)

        # Mapping from renamed dimensions to Scenario dimension names
        MAP = get_reversed_rename_dims()

        filters_to_use = {}
        for dim, values in filters.items():
            if dim in dims:
                # *dim* is in this ixmp object, so the filter can be used
                filters_to_use[MAP.get(dim, dim)] = values

        filters = filters_to_use

    # Retrieve quantity data
    data = getattr(scenario, ix_type)(name, filters)

    # ixmp/GAMS scalar is not returned as pd.DataFrame
    if isinstance(data, dict):  # pragma: no cover
        data = pd.DataFrame.from_records([data])

    # Warn if no values are returned.
    # TODO construct an empty Quantity with the correct dimensions *even if* no values
    #      are returned.
    if len(data) == 0:
        log.debug(f"0 values for {ix_type} {repr(name)} using filters: {repr(filters)}")
        log.debug("May be the cause of subsequent errors.")

    # Convert columns with categorical dtype to str
    data = data.astype(
        {
            dt[0]: str
            for dt in data.dtypes.items()
            if isinstance(dt[1], pd.CategoricalDtype)
        }
    )

    # List of the dimensions
    dims = dims_for_qty(data)

    # Remove the unit from the DataFrame
    try:
        attrs = {"_unit": parse_units(data.pop("unit"))}
    except KeyError:  # pragma: no cover
        # 'equ' are returned without units
        attrs = {}
    except ValueError as e:
        if "mixed units" in e.args[0]:
            # Discard mixed units
            log.warning(f"{name}: {e.args[0]} discarded")
            attrs = {"_unit": registry.Unit("")}
        else:
            # Raise all other ValueErrors
            raise

    # Apply units
    try:
        new_unit = config["units"]["apply"][name]
    except KeyError:
        pass
    else:
        log.info(
            f"{name}: replace units {attrs.get('_unit', '(none)')} with {new_unit}"
        )
        attrs["_unit"] = registry.Unit(new_unit)

    # Set index if 1 or more dimensions
    if len(dims):
        # First rename, then set index
        data = data.rename(columns=RENAME_DIMS).set_index(dims)

    # Convert to a Quantity, assign attrbutes and name
    qty = Quantity(
        data[column], name=name + ("-margin" if column == "mrg" else ""), attrs=attrs
    )

    try:
        # Remove length-1 dimensions for scalars
        qty = qty.squeeze("index", drop=True)
    except (KeyError, ValueError):
        # KeyError if "index" does not exist; ValueError if its length is > 1
        pass

    return qty


def map_as_qty(set_df: pd.DataFrame, full_set):
    """Convert *set_df* to a :class:`.Quantity`.

    For the MESSAGE sets named ``cat_*`` (see :ref:`message_ix:mapping-sets`)
    :meth:`ixmp.Scenario.set` returns a :class:`~pandas.DataFrame` with two columns:
    the *category* set (S1) elements and the *category member* set (S2, also required
    as the argument `full_set`) elements.

    map_as_qty converts such a DataFrame (*set_df*) into a Quantity with two
    dimensions. At the coordinates *(s₁, s₂)*, the value is 1 if *s₂* is mapped from
    *s₁*; otherwise 0.

    A category named 'all', containing all elements of `full_set`, is added
    automatically.

    See also
    --------
    .broadcast_map
    """
    set_from, set_to = set_df.columns
    names = [RENAME_DIMS.get(c, c) for c in set_df.columns]

    # Add an 'all' mapping
    set_df = pd.concat(
        [set_df, pd.DataFrame([("all", e) for e in full_set], columns=set_df.columns)]
    )

    # Add a value column
    set_df["value"] = 1

    return (
        set_df.set_index([set_from, set_to])["value"]
        .rename_axis(index=names)
        .pipe(Quantity)
    )


def update_scenario(scenario, *quantities, params=[]):
    """Update *scenario* with computed data from reporting *quantities*.

    Parameters
    ----------
    scenario : .Scenario
    quantities : .Quantity or pd.DataFrame
        If DataFrame, must be valid input to :meth:`.Scenario.add_par`.
    params : list of str, optional
        For every element of `quantities` that is a pd.DataFrame, the element of
        `params` at the same index gives the name of the parameter to update.
    """
    log.info("Update '{0.model}/{0.scenario}#{0.version}'".format(scenario))
    scenario.check_out()

    for order, (qty, par_name) in enumerate(zip_longest(quantities, params)):
        if not isinstance(qty, pd.DataFrame):
            # Convert a Quantity to a DataFrame
            par_name = qty.name
            new = qty.to_series().reset_index().rename(columns={par_name: "value"})
            new["unit"] = "{:~}".format(qty.attrs["_unit"])  # type: ignore [str-format]
            qty = new

        # Add the data
        log.info(f"  {repr(par_name)} ← {len(qty)} rows")
        scenario.add_par(par_name, qty)

    scenario.commit(f"Data added using {__name__}")



import numpy as np
import pytest

from ixmp.testing import get_cell_output, run_notebook


def test_py_transport(tutorial_path, tmp_path, tmp_env):
    fname = tutorial_path / "transport" / "py_transport.ipynb"
    nb, errors = run_notebook(fname, tmp_path, tmp_env)
    assert errors == []

    # FIXME use get_cell_by_name instead of assuming cell count/order is fixed
    assert np.isclose(get_cell_output(nb, -5)["lvl"], 153.6750030517578)


def test_py_transport_scenario(tutorial_path, tmp_path, tmp_env):
    fname = tutorial_path / "transport" / "py_transport_scenario.ipynb"
    nb, errors = run_notebook(fname, tmp_path, tmp_env)
    assert errors == []

    assert np.isclose(get_cell_output(nb, "scen-z")["lvl"], 153.675)
    assert np.isclose(get_cell_output(nb, "scen-detroit-z")["lvl"], 161.324)


@pytest.mark.rixmp
def test_R_transport(tutorial_path, tmp_path, tmp_env):
    fname = tutorial_path / "transport" / "R_transport.ipynb"
    nb, errors = run_notebook(fname, tmp_path, tmp_env, kernel="IR")
    assert errors == []


@pytest.mark.rixmp
def test_R_transport_scenario(tutorial_path, tmp_path, tmp_env):
    fname = tutorial_path / "transport" / "R_transport_scenario.ipynb"
    nb, errors = run_notebook(fname, tmp_path, tmp_env, kernel="IR")
    assert errors == []



from contextlib import contextmanager
from datetime import datetime, timedelta

import pandas as pd
import pytest
from numpy import testing as npt
from pandas.testing import assert_frame_equal

from ixmp import Scenario, TimeSeries
from ixmp.core import IAMC_IDX
from ixmp.testing import DATA, models

# string columns for timeseries checks
IDX_COLS = ["region", "variable", "unit", "year"]

COLS_FOR_YEARLY_DATA = ["model", "scenario"] + IDX_COLS + ["value"]
COLS_WITH_SUBANNUAL = COLS_FOR_YEARLY_DATA.copy()
COLS_WITH_SUBANNUAL.insert(4, "subannual")


# Utility methods
def expected(df, ts):
    """Modify *df* with the 'model' and 'scenario' name from *ts."""
    return df.assign(model=ts.model, scenario=ts.scenario)


def wide(df):
    """Transform *df* from long to wide format."""
    other_cols = [c for c in df.columns if c not in ["year", "value"] + IAMC_IDX]
    return (
        df.pivot_table(index=IAMC_IDX + other_cols, columns="year", values="value")
        .reset_index()
        .rename_axis(columns=None)
    )


def assert_geodata(obs, exp):
    obs = obs.sort_values("year")
    exp = exp.sort_values("year")
    for column in obs.columns:
        npt.assert_array_equal(exp.get(column), obs.get(column))


def assert_timeseries(scen, exp=DATA["timeseries"], cols=None, subannual=None):
    """Asserts scenario timeseries are similar to expected

    Compares region, variable, unit, year and subannual (if available).
    By default it assumes that datasets are sorted in correct order to compare.

    :param scen:    scenario object
    :param exp:     expected timeseries data
    :param cols:    (optional) column list to sort by
    :param subannual:   passed to 'timeseries()'
    """
    if subannual is None:
        obs = scen.timeseries(region="World")
    else:
        obs = scen.timeseries(region="World", subannual=subannual)

    if cols is not None:
        obs = obs.sort_values(by=cols)
        exp = exp.sort_values(by=cols)
    npt.assert_array_equal(exp[IDX_COLS], obs[IDX_COLS])
    npt.assert_array_almost_equal(exp["value"], obs["value"])
    if "subannual" in exp.columns:
        npt.assert_array_equal(exp[["subannual"]], obs[["subannual"]])


@contextmanager
def transact(ts, condition=True, commit_message=""):
    """Context manager to wrap in a 'transaction'.

    If `condition` is :obj:`True`, the :class:`.TimeSeries`/:class:`.Scenario` `ts` is
    checked out *before* the block begins, and afterwards a commit is made with the
    `commit_message`. If `condition` is :obj:`False`, nothing occurs.
    """
    if condition:
        ts.check_out()
    try:
        yield ts
    finally:
        if condition:
            ts.commit(commit_message)


# Fixtures


class TestTimeSeries:
    """Tests of :class:`.TimeSeries`.

    Since :class:`.Scenario` is a subclass of TimeSeries, all TimeSeries functionality
    should work exactly the same way on Scenario instances. The :meth:`ts` fixture is
    parametrized to yield both TimeSeries and Scenario objects, so every test is run on
    each type.
    """

    @pytest.fixture(scope="function", params=[TimeSeries, Scenario])
    def ts(self, request, mp):
        """An empty TimeSeries with a temporary name on the :func:`mp`."""
        # Use a hash of the pytest node ID to avoid exceeding the maximum
        # length for a scenario name
        node = hash(request.node.nodeid.replace("/", " "))
        # Class of object to yield
        cls = request.param
        yield cls(mp, model=f"test-{node}", scenario="test", version="new")

    # Initialize TimeSeries
    @pytest.mark.parametrize("cls", [TimeSeries, Scenario])
    def test_init(self, test_mp, cls):
        # Something other than a Platform as mp argument
        with pytest.raises(TypeError):
            cls(None, "model name", "scenario name")

        # Invalid version argument
        with pytest.raises(ValueError):
            cls(test_mp, "model name", "scenario name", version=3.4)

    # TimeSeries properties
    def test_default(self, mp, ts):
        # NB this is required before the is_default method can be used
        # FIXME should return False regardless
        ts.commit("")

        # Temporary TimeSeries is has not been set_as_default
        assert not ts.is_default()

        ts.set_as_default()
        assert ts.is_default()

        # NB TimeSeries cannot be cloned, so create a new one with the same
        #    name
        ts2 = TimeSeries(mp, ts.model, ts.scenario, version="new")
        ts2.commit("")
        ts2.set_as_default()

        assert ts2.is_default()
        # Original TimeSeries is no longer default
        assert not ts.is_default()

    def test_run_id(self, ts):
        # New, un-committed TimeSeries has run_id of -1
        assert ts.run_id() == -1

        # The run ID is a positive integer
        ts.commit("")
        assert ts.run_id() > 0 and isinstance(ts.run_id(), int)

    def test_last_update(self, ts):
        # New, un-committed TimeSeries has no last update date
        assert ts.last_update() is None

        ts.commit("")

        # After committing, last_update() returns a string
        last_update = ts.last_update()
        actual = datetime.strptime(last_update, "%Y-%m-%d %H:%M:%S.%f")
        assert abs(actual - datetime.now()) < timedelta(seconds=1)

    @pytest.mark.parametrize("format", ["long", "wide"])
    def test_add_timeseries(self, ts, format):
        data = DATA[0] if format == "long" else wide(DATA[0])

        # Data added
        ts.add_timeseries(data)
        ts.commit("")

        # Error: column 'unit' is missing
        with pytest.raises(ValueError):
            ts.add_timeseries(DATA[0].drop("unit", axis=1))

    @pytest.mark.parametrize("format", ["long", "wide"])
    def test_get(self, ts, format):
        data = DATA[0] if format == "long" else wide(DATA[0])

        ts.add_timeseries(data)
        ts.commit("")

        exp = expected(data, ts)
        args = {}

        if format == "wide":
            args["iamc"] = True

        # Data can be retrieved and has the expected value
        # NB this included check_like=True to be tolerant of JDBCBackend returning
        #    columns in unpredictable order. In pandas 1.2.0, this caused an exception;
        #    see pandas-dev/pandas#39168. Removed until the upstream bug is fixed.
        assert_frame_equal(exp, ts.timeseries(**args))

    @pytest.mark.parametrize("format", ["long", "wide"])
    def test_edit(self, mp, ts, format):
        """Tests that data can be overwritten."""
        data = expected(DATA[0], ts)
        all_data = [data.loc[0:0, :]]
        args = {}

        if format == "wide":
            data = wide(data)
            args["iamc"] = True

        ts.add_timeseries(data)
        ts.commit("initial data")

        # Overwrite existing data
        with transact(ts, commit_message="overwrite existing data"):
            ts.add_timeseries(data)

        df = expected(DATA[2030], ts)
        all_data.append(df)
        if format == "wide":
            df = wide(df)

        # Overwrite and add new values at once
        with transact(ts, commit_message="overwrite and add data"):
            ts.add_timeseries(df)

        # Close and re-open database
        mp.close_db()
        mp.open_db()

        # All four rows are retrieved
        exp = pd.concat(all_data).reset_index(drop=True)
        if format == "wide":
            exp = wide(exp)
        assert_frame_equal(exp, ts.timeseries(**args))

    @pytest.mark.parametrize("cls", [TimeSeries, Scenario])
    def test_edit_with_region_synonyms(self, mp, ts, cls):
        info = dict(model=ts.model, scenario=ts.scenario)

        mp.add_region_synonym("Hell", "World")

        ts.add_timeseries(DATA[0])
        ts.commit("updating timeseries in IAMC format")

        ts = cls(mp, **info)
        assert_frame_equal(expected(DATA[0], ts), ts.timeseries())

        ts.check_out(timeseries_only=True)
        df = wide(DATA[2050]).replace("World", "Hell")
        ts.preload_timeseries()
        ts.add_timeseries(df)
        ts.commit("updating timeseries in IAMC format")

        assert_frame_equal(expected(DATA[2050], ts), ts.timeseries())

    # TODO parametrize format as wide/long
    @pytest.mark.parametrize(
        "commit",
        [
            pytest.param(True),
            pytest.param(
                False,
                marks=pytest.mark.xfail(
                    reason="TimeSeries must be checked in to retrieve data."
                ),
            ),
        ],
    )
    def test_remove(self, mp, ts, commit):
        df = expected(DATA[2050], ts)

        ts.add_timeseries(DATA[2050])
        ts.commit("")

        if not commit:
            ts.check_out()

        assert_frame_equal(df, ts.timeseries())

        # Remove a single data point
        with transact(ts, commit):
            ts.remove_timeseries(df[df.year == 2010])

        # Expected data remains
        exp = df[df.year != 2010].reset_index(drop=True)
        assert_frame_equal(exp, ts.timeseries())

        # Remove two data points
        with transact(ts, commit):
            ts.remove_timeseries(df[df.year.isin([2030, 2050])])

        # Expected data remains
        exp = df[~df.year.isin([2010, 2030, 2050])].reset_index(drop=True)
        assert_frame_equal(exp, ts.timeseries())

        # Remove all remaining data
        with transact(ts, commit):
            ts.remove_timeseries(df)

        # Result is empty
        assert ts.timeseries().empty

    def test_add_geodata(self, ts):
        # Empty TimeSeries includes no geodata
        assert_frame_equal(DATA["geo"].loc[[False, False, False]], ts.get_geodata())

        # Data can be added
        ts.add_geodata(DATA["geo"])
        ts.commit("")

        # Added data can be retrieved
        obs = ts.get_geodata().sort_values("year").reset_index(drop=True)
        assert_frame_equal(DATA["geo"], obs)

    @pytest.mark.parametrize(
        "rows", [(1,), (1, 2), (0, 1, 2)], ids=["single", "multiple", "all"]
    )
    def test_remove_geodata(self, ts, rows):
        ts.add_geodata(DATA["geo"])
        ts.remove_geodata(DATA["geo"].take(rows))
        ts.commit("")

        mask = [i not in rows for i in range(len(DATA["geo"]))]
        # Expected rows have been removed
        exp = DATA["geo"].iloc[mask].reset_index(drop=True)
        obs = ts.get_geodata().sort_values("year").reset_index(drop=True)
        assert_frame_equal(exp, obs)

    @pytest.mark.parametrize("format", ["long", "wide"])
    @pytest.mark.parametrize(
        "N",
        (
            256,
            # This fails at commit(), not at add_timeseries()
            pytest.param(
                257,
                marks=pytest.mark.xfail(raises=RuntimeError),
            ),
        ),
    )
    def test_long_variable_name(self, ts, format, N):
        """Variable names up to 256 characters can be added or removed."""
        data = (DATA[0] if format == "long" else wide(DATA[0])).copy()

        # Use long variable name, max 256 characters
        data.variable = "x" * N

        ts.add_timeseries(data)
        ts.commit("")

        data = ts.timeseries()

        with transact(ts):
            ts.remove_timeseries(data)

    # FIXME all tests below this line need cleanup

    @pytest.mark.parametrize("format", ["long", "wide"])
    def test_add_timeseries_with_extra_col(self, caplog, ts, format):
        _data = DATA[0].copy()
        _data["climate_model"] = [0, 1]
        data = _data if format == "long" else wide(_data)

        # check that extra column wasn't dropped by `wide(_data)
        assert "climate_model" in data.columns

        # Data added
        ts.add_timeseries(data)
        # TODO: add check that warning message is displayed
        ts.commit("")
        assert [
            "Dropped extra column(s) ['climate_model'] from data"
        ] == caplog.messages

    def test_new_timeseries_as_iamc(self, test_mp):
        # TODO rescue use of subannual= here

        scen = TimeSeries(test_mp, *models["h2g2"], version="new", annotation="fo")
        scen.add_timeseries(
            DATA["timeseries"].pivot_table(values="value", index=IDX_COLS)
        )
        scen.commit("importing a testing timeseries")

        # compare returned dataframe - default behaviour set to 'auto'
        assert_timeseries(scen)
        # test behaviour of 'auto' explicitly
        assert_timeseries(scen, subannual="auto")
        # test behaviour of 'False' explicitly
        assert_timeseries(scen, subannual=False)

        # test behaviour of 'True' explicitly
        exp = (
            DATA["timeseries"].pivot_table(values="value", index=IDX_COLS).reset_index()
        )
        exp["model"] = "Douglas Adams"
        exp["scenario"] = "Hitchhiker"
        exp["subannual"] = "Year"
        assert_timeseries(scen, exp=exp, cols=COLS_WITH_SUBANNUAL, subannual=True)

    def test_new_timeseries_error(self, test_mp):
        scen = TimeSeries(test_mp, *models["h2g2"], version="new", annotation="fo")
        df = {"year": [2010, 2020], "value": [23.5, 23.6]}
        df = pd.DataFrame.from_dict(df)
        df["region"] = "World"
        df["variable"] = "Testing"
        # column `unit` is missing
        pytest.raises(ValueError, scen.add_timeseries, df)

    def test_new_subannual_timeseries_as_iamc(self, mp):
        mp.add_timeslice("Summer", "Season", 1.0 / 4)
        scen = TimeSeries(mp, *models["h2g2"], version="new", annotation="fo")
        timeseries = DATA["timeseries"].pivot_table(values="value", index=IDX_COLS)
        scen.add_timeseries(timeseries)
        scen.commit("adding yearly data")

        # add subannual timeseries data
        ts_summer = timeseries.copy()
        ts_summer["subannual"] = "Summer"
        scen.check_out()
        scen.add_timeseries(ts_summer)
        scen.commit("adding subannual data")

        # generate expected dataframe+
        ts_year = timeseries.copy()
        ts_year["subannual"] = "Year"
        exp = pd.concat([ts_year, ts_summer]).reset_index()
        exp["model"] = "Douglas Adams"
        exp["scenario"] = "Hitchhiker"

        # compare returned dataframe - default behaviour set to 'auto'
        assert_timeseries(scen, exp=exp[COLS_WITH_SUBANNUAL], cols=COLS_WITH_SUBANNUAL)
        # test behaviour of 'auto' explicitly
        assert_timeseries(
            scen,
            exp=exp[COLS_WITH_SUBANNUAL],
            cols=COLS_WITH_SUBANNUAL,
            subannual="auto",
        )
        # test behaviour of 'True' explicitly
        assert_timeseries(
            scen, exp=exp[COLS_WITH_SUBANNUAL], cols=COLS_WITH_SUBANNUAL, subannual=True
        )
        # setting False raises an error because subannual data exists
        pytest.raises(ValueError, scen.timeseries, subannual=False)

    def test_fetch_empty_geodata(self, mp):
        scen = TimeSeries(mp, *models["h2g2"], version="new", annotation="fo")
        empty = scen.get_geodata()
        assert_geodata(empty, DATA["geo"].loc[[False, False, False]])

    def test_remove_multiple_geodata(self, mp):
        scen = TimeSeries(mp, *models["h2g2"], version="new", annotation="fo")
        scen.add_geodata(DATA["geo"])
        row = DATA["geo"].loc[[False, True, True]]
        scen.remove_geodata(row)
        scen.commit("adding geodata (references to map layers)")
        assert_geodata(scen.get_geodata(), DATA["geo"].loc[[True, False, False]])



import logging

from pkg_resources import DistributionNotFound, get_distribution

from ._config import config
from .backend import BACKENDS, ItemType
from .backend.jdbc import JDBCBackend
from .core import IAMC_IDX, Platform, Scenario, TimeSeries
from .model import MODELS
from .model.base import ModelError
from .model.dantzig import DantzigModel
from .model.gams import GAMSModel
from .reporting import Reporter
from .utils import show_versions

__all__ = [
    "IAMC_IDX",
    "ItemType",
    "ModelError",
    "Platform",
    "Reporter",
    "Scenario",
    "TimeSeries",
    "config",
    "log",
    "show_versions",
]

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:  # pragma: no cover
    # Package is not installed
    __version__ = "999"

# Register Backends provided by ixmp
BACKENDS["jdbc"] = JDBCBackend

# Register Models provided by ixmp
MODELS.update(
    {
        "default": GAMSModel,
        "gams": GAMSModel,
        "dantzig": DantzigModel,
    }
)


# Configure the 'ixmp' logger: write messages to stdout, defaulting to level WARNING
# and above
log = logging.getLogger(__name__)
log.addHandler(logging.StreamHandler())
log.setLevel(logging.WARNING)



import json
import logging
import os
from copy import copy, deepcopy
from pathlib import Path
from typing import Dict

log = logging.getLogger(__name__)


class _JSONEncoder(json.JSONEncoder):
    """Helper for writing config to file.

    The default JSONEncoder does not automatically convert pathlib.Path objects.
    """

    def default(self, o):
        if isinstance(o, Path):
            return str(o)
        return json.JSONEncoder.default(self, o)


def _iter_config_paths():
    """Yield recognized configuration paths, in order of priority."""
    try:
        yield "environment (IXMP_DATA)", Path(os.environ["IXMP_DATA"]).resolve()
    except KeyError:
        pass

    try:
        yield "environment (XDG_DATA_HOME)", Path(
            os.environ["XDG_DATA_HOME"], "ixmp"
        ).resolve()
    except KeyError:
        pass

    yield "default", Path.home().joinpath(".local", "share", "ixmp")


#: Registered configuration keys; name -> (type, default value).
KEYS = {
    "platform": (
        dict,
        {
            "default": "local",
            "local": {
                "class": "jdbc",
                "driver": "hsqldb",
                "path": next(_iter_config_paths())[1].joinpath("localdb", "default"),
            },
        },
    ),
}


def _locate(filename=None):
    """Locate an existing `filename` in the ixmp config directories.

    If `filename` is :obj:`None` (the default), only directories are located.
    """
    tried = []
    for label, directory in _iter_config_paths():
        # Locate either a specific file or an existing directory
        path = directory.joinpath(filename) if filename else directory

        if path.exists():
            return path.resolve()

        tried.append(str(directory))

    raise FileNotFoundError(
        "Could not find " + (f"{filename} in " if filename else "any of ") + repr(tried)
    )


class Config:
    """Configuration for ixmp.

    Config stores two kinds of data: simple keys with a single value, and structured
    Platform information.

    ixmp has no built-in simple keys; however, it can store keys for other packages
    that build on ixmp, such as :mod:`message_ix`.

    Parameters
    ----------
    read : bool
        Read ``config.json`` on startup.
    """

    #: Full-resolved path of the ``config.json`` file.
    path = None

    # Configuration values
    values: Dict[str, object] = dict()

    def __init__(self, read=True):
        # Default values
        self.clear()

        # Read configuration from file; store the path at which it was located
        if read:
            self.read()

    def read(self):
        """Try to read configuration keys from file.

        If successful, the attribute :attr:`path` is set to the path of the file.
        """
        try:
            # Locate and read the configuration file
            config_path = _locate("config.json")
        except FileNotFoundError:
            contents = "{}"
        else:
            self.path = config_path.resolve()
            contents = config_path.read_text()

        try:
            # Parse JSON and set values; _strict=False tolerates unregistered values
            for key, value in json.loads(contents).items():
                self.set(key, value, _strict=False)
        except json.JSONDecodeError:
            print(config_path, contents)
            raise

    # Public methods

    def get(self, name):
        """Return the value of a configuration key `name`."""
        return self.values[name]

    def register(self, name, type_, default=None):
        """Register a new configuration key.

        Parameters
        ----------
        name : str
            Name of the new key; must not already exist.
        type_ : object
            Type of the key's value, such as :obj:`str` or :class:`pathlib.Path`.
        default : any, optional
            Default value for the key. If not supplied, the `type` is called to supply
            the default value, e.g. ``str()``.
        """
        if name in KEYS:
            raise KeyError(f"configuration key {repr(name)} already defined")

        # Register the key for future clear()
        KEYS[name] = (type_, default)

        # Also set on the current config object
        self.values[name] = default or type_()

    def unregister(self, name):
        """Unregister and clear the configuration key `name`."""
        KEYS.pop(name, None)
        self.values.pop(name, None)

    def set(self, name, value, _strict=True):
        """Set configuration key `name` to `value`."""
        if value is None:
            return

        # Retrieve the type for `name`; or None if unregistered
        type_ = KEYS.get(name, (None,))[0]

        if type_ or _strict:
            try:
                # Attempt to cast to the correct type
                value = type_(value)
            except TypeError:
                # _strict and unregistered key; tried to call None(value)
                raise KeyError(name)
            except Exception:
                raise TypeError(
                    f"expected {type_} for {repr(name)}; got {type(value)} "
                    f"{repr(value)}"
                )

        self.values[name] = value

    def clear(self):
        """Clear all configuration keys by setting empty or default values.

        :meth:`clear` also sets the default local platform::

          {
            "platform": {
              "default": "local",
              "local": {
                "class": "jdbc",
                "driver": "hsqldb",
                "path": "~/.local/share/ixmp/localdb/default"
              },
          }
        """
        self.values = dict()
        for name, (type_, default) in KEYS.items():
            self.values[name] = default or type_()

        # Set the default local database path; changed versus KEYS if IXMP_DATA has been
        # altered since the module was imported
        local = next(_iter_config_paths())[1].joinpath("localdb", "default")
        self.values["platform"]["local"]["path"] = local

    def save(self):
        """Write configuration keys to file.

        ``config.json`` is created in the first of the ixmp configuration directories
        that exists. Only non-null values are written.
        """
        # Use the first identifiable path
        _, config_dir = next(_iter_config_paths())
        path = config_dir / "config.json"

        # TODO merge with existing configuration

        # Make the directory to contain the configuration file
        path.parent.mkdir(parents=True, exist_ok=True)

        values = deepcopy(self.values)
        for key, type_ in KEYS.items():
            # Don't store empty strings
            if type_ is str and values[key] == "":
                values.pop(key)

        # Write the file
        log.info("Updating configuration file: {}".format(path))
        path.write_text(_JSONEncoder(indent=2).encode(values))

        # Update the path attribute to match the written file
        self.path = path

    def add_platform(self, name, *args, **kwargs):
        """Add or overwrite information about a platform.

        Parameters
        ----------
        name : str
            New or existing platform name.
        args
            Positional arguments. If `name` is 'default', `args` must be a single
            string: the name of an existing configured Platform. Otherwise, the first
            of `args` specifies one of the :obj:`~.BACKENDS`, and the remaining `args`
            differ according to the backend.
        kwargs
            Keyword arguments. These differ according to backend.

        See also
        --------
        Backend.handle_config
        JDBCBackend.handle_config
        """
        args = list(args)
        if name == "default":
            assert len(args) == 1
            info = args[0]

            if info not in self.values["platform"]:
                raise ValueError(f"Cannot set unknown {repr(info)} as default platform")
        else:
            from ixmp.backend import BACKENDS

            try:
                # Get the backend class
                cls = args.pop(0)
                backend_class = BACKENDS[cls]
            except IndexError:
                raise ValueError("Must give at least 1 arg: backend class")
            except KeyError:
                raise ValueError(f"No backend named {repr(cls)}")

            # Use the backend class' method to handle the arguments
            info = backend_class.handle_config(args, kwargs)

            info.setdefault("class", cls)

        if name in self.values["platform"]:
            log.warning(
                "Overwriting existing config: " + repr(self.values["platform"][name])
            )

        self.values["platform"][name] = info

    def get_platform_info(self, name):
        """Return information on configured Platform *name*.

        Parameters
        ----------
        name : str
            Existing platform. If `name` is 'default', then the information for the
            default platform is returned.

        Returns
        -------
        dict
            The 'class' key specifies one of the :obj:`~.BACKENDS`. Other keys vary by
            backend class.

        Raises
        ------
        KeyError
            If `name` is not configured as a platform.
        """
        if name == "default":
            # The 'default' key stores the name of another config'd platform
            name = self.values["platform"].get(name, None)
        try:
            return name, copy(self.values["platform"][name])
        except KeyError:
            raise ValueError(
                f"platform name {repr(name)} not among "
                + repr(sorted(self.values["platform"].keys()))
                + f"\nfrom {self.path}"
            )

    def remove_platform(self, name):
        """Remove the configuration for platform `name`."""
        self.values["platform"].pop(name)


#: Default :mod:`ixmp` configuration object.
config = Config()
